#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Stimulus object for drawing arbitrary bitmap carriers with an arbitrary
second order envelope carrier and envelope can vary independently for
orientation, frequencyand phase. Also does beat stimuli. """

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# some code provided by Andrew Schofield
# Distributed under the terms of the GNU General Public License (GPL).

# Requires shaders if you don't have them it will just throw and error.
# Ensure setting pyglet.options['debug_gl'] to False is done prior to any
# other calls to pyglet or pyglet submodules, otherwise it may not get picked
# up by the pyglet GL engine and have no effect.
# Shaders will work but require OpenGL2.0 drivers AND PyOpenGL3.0+

import pyglet
pyglet.options['debug_gl'] = False
import ctypes
GL = pyglet.gl
try:
    from PIL import Image
except ImportError:
    import Image
import psychopy  # so we can get the __path__
from psychopy import logging
from psychopy.visual import filters
from psychopy.tools.arraytools import val2array
from psychopy.tools.attributetools import attributeSetter
from .grating import GratingStim
import numpy
from numpy import exp, sin, cos
from numpy.fft import fft2, ifft2, fftshift, ifftshift

from . import shaders as _shaders


class NoiseStim(GratingStim):
    """A stimulus with 2 textures: a radom noise sample and a mask

    **Example**::

        noise1 = noise = visual.NoiseStim(
                        win=win, name='noise',units='pix',
                        noiseImage='testImg.jpg', mask='circle',
                        ori=1.0, pos=(0, 0), size=(512, 512), sf=None, phase=0,
                        color=[1,1,1], colorSpace='rgb', opacity=1, blendmode='add', contrast=1.0,
                        texRes=512, filter='None', imageComponent='Phase'
                        noiseType='Gabor', noiseElementSize=4, noiseBaseSf=32.0/512,
                        noiseBW=1.0, noiseBWO=30, noiseFractalPower=-1,noiseFilterLower=3/512, noiseFilterUpper=8.0/512.0,
                        noiseFilterOrder=3.0, noiseClip=3.0, filter=False, interpolate=False, depth=-1.0)
        # gives a circular patch of noise made up of scattered Gabor elements with peak frequency = 32.0/512 cycles per pixel,
        # orientation = 0 , frequency bandwidth = 1 octave and orientation bandwidth 30 degrees

    **Types of noise available**

    * Binary, Normal, Uniform - pixel based noise samples drawn from a binary (blank and white), normal or uniform distribution respectively. Binary noise is always exactly zero mean, Normal and Uniform are approximately so.
      Parameters:

        * noiseElementSize - (can be a tuple) defines the size of the noise elements in the components units.
        * noiseClip the values in normally distributed noise are divided by noiseClip to limit excessively high or low values.
          However, values can still go out of range -1 to 1 whih will throw a soft error message high values of noiseClip are recommended if using 'Normal'
    
    * **Gabor**, **Isotropic**: Effectively a dense scattering of Gabor elements with random amplitude and fixed orientation
      for Gabor or random orientation for Isotropic noise. In practice the desired amplitude spectrum for the noise is
      built in Fourier space with a random phase spectrum. DC term is set to zero - ie zero mean.
      Parameters:

        * noiseBaseSf - centre spatial frequency in the component units.
        * noiseBW - spatial frequency bandwidth full width half height in octaves.
        * ori - center orientation for Gabor noise (works as for gratingStim so twists the final image at render time).
        * noiseBWO - orientation bandwidth for Gabor noise full width half height in degrees.
        * noiseOri - alternative center orientation for Gabor which sets the orientation during the image build rather
          than at render time. Useful for setting the orientation of a filter to be applied to some other noise type
          with a different base orientation.


    * **Filtered** - A white noise sample that has been filtered with a low, high or bandpass Butterworth filter. The
      initial sample can have its spectrum skewed towards low or high frequencies. The contrast of the noise falls
      by half its maximum (3dB) at the cutoff frequencies.
      Parameters:

        * noiseFilterUpper - upper cutoff frequency - if greater than texRes/2 cycles per image low pass filter used.
        * noiseFilterLower - Lower cutoff frequency - if zero low pass filter used.
        * noiseFilterOrder - The order of the filter controls the steepness of the falloff outside the passband is
          zero no filter is applied.
        * noiseFractalPower - spectrum = f^noiseFractalPower  - determines the spatial frequency bias of the initial
          noise sample. 0 = flat spectrum, negative = low frequency bias, positive = high frequency bias, -1 = fractal
          or brownian noise.
        * noiseClip - determines clipping values and rescaling factor such that final rms contrast is close to that
          requested by contrast parameter while keeping pixel values in range -1, 1.

    * **White** - A short cut to obtain noise with a flat, unfiltered spectrum. In practice the desired amplitude spectrum
      is built in the Fourier Domain with a random phase spectrum. DC term is set to zero - ie zero mean
      Note despite name the noise contains all grey levels.
      Parameters:

        * noiseClip - determines clipping values and rescaling factor such that final rms contrast is close to that
          requested by contrast parameter while keeping pixel values in range -1, 1.

    * **Image**: A noise sample whose spatial frequency spectrum is taken from the supplied image. In practice the
      desired amplitude spectrum is taken from the image and paired with a random phase spectrum. DC term is set to
      zero - ie zero mean.
      Parameters:

        * noiseImage name of ndarray or image file from which to take spectrum - should be same size as largest side
          requested for component if units is pix or texRes x texRes otherwise
        * imageComponent: 'Phase' randomizes the phase spectrum leaving the amplitude spectrum untouched. 'Amplitude'
          randomizes the amplitude spectrum leaving the phase spectrum untouched - retains spatial structure of image.
          'Neither' keeps the image as is - but you can now apply a spatial filter to the image.
        * noiseClip - determines clipping values and rescaling factor such that final rms contrast is close to that
          requested by contrast parameter while keeping pixel values in range -1, 1.

    **Filter parameter**

    * Butterworth: a spectral filter defined by the filtered noise parameters will be applied to the other noise types.
    * Gabor: a spectral filter defined by the Gabor noise parameters will be applied to the other noise types.
    * Isotropic: then a spectral filter defined by the Isotropic noise parameters will be applied to the other noise
      types.
    
    
    **Updating noise samples and timing**

    The noise is rebuilt at next call of the draw function whenever a parameter starting 'noise' is notionally changed
    even if the value does not actually change every time. eg. setting a parameter to update every frame will cause a
    new noise sample on every frame but see below.
    A rebuild can also be forced at any time using the buildNoise() function.
    The updateNoise() function can be used at any time to produce a new random saple of noise without doing a full
    build. ie it is quicker than a full build.
    Both buildNoise and updateNoise can be slow for large samples. 
    Samples of Binary, Normal or Uniform noise can usually be made at frame rate using noiseUpdate. 
    Updating or building other noise types at frame rate may result in dropped frames. 
    An alternative is to build a large sample of noise at the start of the routien and place it off the screen then
    cut a samples out of this at random locations and feed that as a numpy array into the texture of a visible
    gratingStim.

    **Notes on size**
    If units = pix and noiseType = Binary, Normal or Uniform will make noise sample of requested size.
    If units = pix and noiseType is Gabor, Isotropic, Filtered, White, Coloured or Image will make square noise sample with side length equal that of the largest dimetions requested.
    if units is not pix will make square noise sample with side length equal to texRes then rescale to present.
    
    **Notes on interpolation**
    For pixel based noise interpolation = nearest is usually best.
    For other noise types linear is better if the size of the noise sample does not match the final size of the image well.
    
    **Notes on frequency**
    Frequencies for cutoffs etc are converted between units for you but can be counter intuitive. 1/size is always 1 cycle per image.
    For the sf (final spatial frequency) parameter itself 1/size (or None for units pix) will faithfully represent the image without further scaling.
        
    Filter cuttoff and Gabor/Isotropic base frequencies should not be too high you should aim to keep them below 0.5 c/pixel on the screen.
    The function will produce an error when it can't draw the stimulus in the buffer but it may still be wrong when displayed.
    
    **Notes on orientation and phase**
    The ori parameter twists the final image so the samples in noiseType Binary, Normal or Uniform will no longer be aligned to the sides of the monitor if ori is not a multiple of 90.
    Most other noise types look broadly the same for all values of ori but the specific sample shown can be made to rotate by changing ori.
    The dominant orientation for Gabor noise is determined by ori at render time, not before.
    
    The phase parameter similarly shifts the sample around within the display window at render time and will not choose new random phases for the noise sample.
    """

    def __init__(self,
                 win,
                 mask="none",
                 units="",
                 pos=(0.0, 0.0),
                 size=None,
                 sf=None,
                 ori=0.0,
                 phase=(0.0, 0.0),
                 noiseType=None,
                 noiseElementSize=16,
                 noiseBaseSf=1,
                 noiseBW=1,
                 noiseBWO=30,
                 noiseOri=0,
                 noiseFractalPower=0.0,
                 noiseFilterUpper=50,
                 noiseFilterLower=0,
                 noiseFilterOrder=0.0,
                 noiseClip=1,
                 noiseImage=None,
                 imageComponent='Phase',
                 filter=None,
                 texRes=128,
                 rgb=None,
                 dkl=None,
                 lms=None,
                 color=(1.0, 1.0, 1.0),
                 colorSpace='rgb',
                 contrast=0.5,  # see doc
                 opacity=1.0,
                 depth=0,
                 rgbPedestal=(0.0, 0.0, 0.0),
                 interpolate=False,
                 blendmode='avg',
                 name=None,
                 autoLog=None,
                 autoDraw=False,
                 maskParams=None):
        """ """  # Empty docstring. All doc is in attributes
        # what local vars are defined (these are the init params) for use by
        # __repr__

        self._initParams = dir()

        if noiseType is None:
            msg = ('noiseType not recognized. Valid types are: \n'
                   'binary, uniform, normal, white, filtered, gabor, '
                   'isotropic, or image.')
            raise ValueError(msg)
        elif noiseType == 'image' and noiseImage is None:
            msg = ('You need to supply an image via the noiseImage keyword '
                   'argument.')
            raise ValueError(msg)

        for unecess in ['self', 'rgb', 'dkl', 'lms']:
            self._initParams.remove(unecess)
        # initialise parent class
        GratingStim.__init__(self, win, tex='sin',
                             units=units, pos=pos, size=size, sf=sf,
                             ori=ori, phase=phase,
                             color=color, colorSpace=colorSpace,
                             contrast=contrast, opacity=opacity,
                             depth=depth, interpolate=interpolate,
                             name=name, autoLog=autoLog, autoDraw=autoDraw,
                             blendmode=blendmode,
                             maskParams=None)

        # UGLY HACK: Some parameters depend on each other for processing.
        # They are set "superficially" here.
        # TO DO: postpone calls to _createTexture, setColor and
        # _calcCyclesPerStim whin initiating stimulus

        self.__dict__['mask'] = mask
        self.__dict__['maskParams'] = maskParams

        self.blendmode=blendmode
        self.mask = mask
        self.texRes=int(texRes)
        self.noiseType=noiseType
        self.noiseImage=noiseImage
        self.imageComponent=imageComponent
        self.noiseElementSize=noiseElementSize
        self.noiseBaseSf=float(noiseBaseSf)
        self.noiseBW=float(noiseBW)
        self.noiseBWO=float(noiseBWO)
        self.noiseOri=float(noiseOri)
        self.noiseFractalPower=float(noiseFractalPower)
        self.noiseFilterUpper=float(noiseFilterUpper)
        self.noiseFilterLower=float(noiseFilterLower)
        self.noiseFilterOrder=float(noiseFilterOrder)
        if noiseClip != 'none':
            self.noiseClip=float(noiseClip)
        else:
            self.noiseClip=noiseClip
        self.filter = filter
        self.local = numpy.ones((texRes, texRes), dtype=numpy.ubyte)
        self.local_p = self.local.ctypes
        self._sideLength=1.0   
        self._mysize=512         # in unlikely case where it does not get set anywhere else before use.
        self.buildNoise()
        self._needBuild = False


    @attributeSetter
    def noiseType(self, value):
        """Type of noise to generate
            'Binary, Normal and Uniform' produce pixel based random samples from the indicated distribitions. 
                Binary has zero mean lumiannce, Normal and Uniform approximate this.
            'Gabor and Isotropic' produce dense, random scatterd Gabor elements with 
                fixed (Gabor) or random (Isotropic) orientations. Zero mean lumiannce.
            'White' produces white noise with a flat amplitude spectrum. Zero mean lumiannce.
            'Filtered' Produces white noise filtered by a low-, high- or band-pass filter. Zero mean lumiannce.
                Use the noiseFractalPower attribute to skew the spectrum of filtered noise.
            'Image' produces noise with the same spectrum as the supplied image but with mean lumiance set to zero.
        
        """
        self.__dict__['noiseType'] = value
        self._needUpdate = True
        self._needBuild = True
        
    @attributeSetter
    def noiseImage(self, value):
        """Image from which to derive the amplitude or phase spectrum of noise type Image. 
        """
       
        self.__dict__['noiseImage'] = value
        self._needUpdate = True
        self._needBuild = True
        
    @attributeSetter
    def imageCompoment(self, value):
        """Which component of an image to randomise, amplitude or phase. 
        """
       
        self.__dict__['imageComponent'] = value
        self._needUpdate = True
        self._needBuild = True

    @attributeSetter
    def noiseElementSize(self, value):
        """Noise element size for Binary, Normal or Uniform noise.
           In the units of the stimulus.
        """
        
        self.__dict__['noiseElementSize'] = value
        self._needUpdate = True
        self._needBuild = True
        
    @attributeSetter
    def noiseBaseSf(self, value):
        """Spatial frequency for Gabor or Isotropic noise in cycles per unit.
           Eg c/deg if units = degrees.
        """
        
        self.__dict__['noiseBaseSf'] = value
        self._needUpdate = True
        self._needBuild = True
        
    @attributeSetter
    def noiseBW(self, value):
        """Spatial frequency bandwidth for Gabor or Isotropic noise, full width at half height in octaves.
        """
        
        self.__dict__['noiseBW'] = value
        self._needUpdate = True 
        self._needBuild = True
        
    @attributeSetter
    def noiseBWO(self, value):
        """Orientation bandwidth for Gabor noise, full width at half height in degrees
        """
        
        self.__dict__['noiseBWO'] = value
        self._needUpdate = True
        self._needBuild = True
        
    @attributeSetter
    def noiseFractalPower(self, value):
        """Exponent for 'coloured' noise. 
           Amplitide spectrum = f^noiseFractalPower.
           -1 gives pink noise.
           Note power spectrum of a pink noise image should
           fall as f^-2. But as power spectrum = amplitude spectrum squared
           this is achieved by setting amplitude spectrum to f^-1.
        """
        
        self.__dict__['noiseFractalPower'] = value
        self._needUpdate = True
        self._needBuild = True
        
    @attributeSetter
    def noiseFilterUpper(self, value):
        """Upper cuttoff for filtered noise. 
           In cycles/unit eg c/deg when units is degrees.
           > size/2 creates high pass filter.   
        """
        
        self.__dict__['noiseFilterUpper'] = value
        self._needUpdate = True
        self._needBuild = True
        
    @attributeSetter
    def noiseFilterLower(self, value):
        """Lower cuttoff for filtered noise. 
           In cycles/unit eg c/deg when units is degrees.
           Zero creates low pass filter.
        """
        
        self.__dict__['noiseFilterLower'] = value
        self._needUpdate = True
        self._needBuild = True
        
    @attributeSetter
    def noiseFilterOrder(self, value):
        """Order of Butterworth filter for filtered noise. High = fast fall off withfrequency outside pass band.
        """
        
        self.__dict__['noiseFilterOrder'] = value
        self._needUpdate = True
        self._needBuild = True
        
    @attributeSetter
    def noiseClip(self, value):
        """Ignored for types 'Binary and Uniform'.
            For 'Normal' noise pixel values are divided by noiseClip 
            to limit the standard deviation of the noise values.
            
            For all other noise types noiseClip determines the 
            level at which pixel values are cliped and subsequently 
            re-scaled so as to produce a final image appraching the 
            desired RMS contrast.
            High values will tend to reduce the ultimate RMS 
            contrast but increase fidelity of appearance.
            Low values prioritise accurate final contrast
            but result in a binarised or thresholded appearance.
            
            noiseClip is used to scale the luminance values as
            above whenever a filter is applied to the noise sample, 
            regardless of the initial type of noise requested. 
        """
        
        self.__dict__['noiseClip'] = value
        self._needUpdate = True
        self._needBuild = True

    @attributeSetter
    def filter(self, value):
        """If True apply spatial frequency filter to noise."""
        
        self.__dict__['filter'] = value
        self._needUpdate = True
        self._needBuild = True

    @attributeSetter
    def texRes(self, value):
        """Power-of-two int. Sets the resolution of the mask and texture.
        maybe used as the side length for generating the noise texture but is not used if units = pix.

        :ref:`Operations <attrib-operations>` supported.
        """
        self.__dict__['texRes'] = value

        # ... now rebuild textures (call attributeSetters without logging).
        
        self._needUpdate = True 
        self._needBuild = True
        
        if hasattr(self, 'tex'):
            self._set('tex', self.tex, log=False)
        if hasattr(self, 'mask'):
            self._set('mask', self.mask, log=False)

    def setTexRes(self, value, log=None):
        self._set('texRes', value, log=log)
        
    def setNoiseType(self, value, log=None):
        self._set('noiseType', value, log=log)
        
    def setNoiseImage(self, value, log=None):
        self._set('noiseImage', value, log=log)
    
    def setImageCompoment(self, value, log=None):
        self._set('imageCompoment',value, log=log)
    
    def setNoiseClip(self, value, log=None):
        self._set('noiseClip', value, log=log)
        
    def setFilter(self, value, log=None):
        self._set('filter', value, log=log)
        
    def setNoiseElementSize(self, value, log=None):
        self._set('noiseElementSize', value, log=log)
        
    def setNoiseBaseSf(self, value, log=None):
        self._set('noiseBaseSf', value, log=log)
        
    def setNoiseBW(self, value, log=None):
        self._set('noiseBW', value, log=log)
        
    def setNoiseBWO(self, value, log=None):
        self._set('noiseBWO', value, log=log)
       
    def setNoiseFractalPower(self, value, log=None):
        self._set('noiseFractalPower', value, log=log)
       
    def setNoiseFilterUpper(self, value, log=None):
        self._set('noiseFilterUpper', value, log=log)
        
    def setNoiseFilterLower(self, value, log=None):
        self._set('noiseFilterLower', value, log=log)
        
    def setNoiseFilterOrder(self, value, log=None):
        self._set('noiseFilterOrder', value, log=log)

    def setblendMode(self, value, log=None):
        self._set('blendMode', value, log=log)

    def draw(self, win=None):
        """Draw the stimulus in its relevant window. You must call
        this method after every MyWin.flip() if you want the
        stimulus to appear on that frame and then update the screen
        again.
        """
        if win is None:
            win = self.win
        saveBlendMode = win.blendMode
        win.setBlendMode(self.blendmode, log=False)
        self._selectWindow(win)

        #do scaling
        GL.glPushMatrix()  # push before the list, pop after
        win.setScale('pix')
        #the list just does the texture mapping
        GL.glColor4f(*self._foreColor.render('rgba1'))

        # re-build the noise if not done so since last parameter update
        if self._needBuild:
            self.buildNoise()
        # remake textures if necessary
        if self._needTextureUpdate:
            self.setTex(value=self.tex, log=False)
        if self._needUpdate:
            self._updateList()
        GL.glCallList(self._listID)

        #return the view to previous state
        GL.glPopMatrix()
        win.setBlendMode(saveBlendMode, log=False)
        
    def _filter(self, FT):
        """ Helper function to apply Butterworth filter in 
            frequensy domain.
        """
        filterSize = numpy.max(self._mysize)
        pin=filters.makeRadialMatrix(matrixSize=filterSize, center=(0,0), radius=1.0)
        pin[int(filterSize / 2)][int(filterSize / 2)] = 0.00000001  # Prevents divide by zero error. This is DC and is set to zero later anyway.
        FT = numpy.multiply(FT,(pin) ** self.noiseFractalPower)
        if self.noiseFilterOrder > 0.01:
            if self._upsf<(filterSize/2.0):
                filter = filters.butter2d_lp_elliptic(size = [filterSize,filterSize], 
                                                            cutoff_x = self._upsf / filterSize, 
                                                            cutoff_y = self._upsf / filterSize, 
                                                            n=self.noiseFilterOrder, 
                                                            alpha=0, 
                                                                offset_x = 0.5/filterSize,  #because FFTs are slightly off centred.
                                                                offset_y = 0.5/filterSize)
            else:
                filter = numpy.ones((int(filterSize),int(filterSize)))
            if self._lowsf > 0:
                if self._lowsf > filterSize/2:
                    msg = ('Lower cut off frequency for filtered '
                    'noise is too high (exceeds Nyquist limit).')
                    raise Warning(msg)
                filter = filter-filters.butter2d_lp_elliptic(size = [filterSize,filterSize], 
                                                                cutoff_x = self._lowsf / filterSize, 
                                                                cutoff_y = self._lowsf / filterSize, 
                                                                n = self.noiseFilterOrder, 
                                                                alpha = 0, 
                                                                offset_x = 0.5/filterSize, #because FFTs are slightly off centred.
                                                                offset_y = 0.5/filterSize)
            return FT * filter
        else:
            return FT
            
    def _isotropic(self, FT):
        """ Helper function to apply isotropic filter in 
            frequensy domain.
        """
        if self._sf > self._mysize / 2:
            msg = ('Base frequency for isotropic '
                  'noise is  too high (exceeds Nyquist limit).')
            raise Warning(msg)
        localf = self._sf / self._mysize
        linbw = 2 ** self.noiseBW
        lowf = 2.0 * localf / (linbw+1.0)
        highf = linbw * lowf
        FWF = highf - lowf
        sigmaF = FWF / (2*numpy.sqrt(2*numpy.log(2)))
        pin = filters.makeRadialMatrix(matrixSize=self._mysize, center=(0,0), radius=2)
        filter = filters.makeGauss(pin, mean=localf, sd=sigmaF)
        return FT*filter
        
    def _gabor(self, FT):
        """ Helper function to apply Gabor filter in 
            frequensy domain.
        """
        if self._sf > self._mysize / 2:
            msg = ('Base frequency for Gabor '
                  'noise is  too high (exceeds Nyquist limit).')
            raise Warning(msg)
        localf = self._sf / self._mysize
        linbw = 2 ** self.noiseBW
        lowf = 2.0 * localf / (linbw + 1.0)
        highf = linbw * lowf
        FWF = highf - lowf
        sigmaF = FWF/(2*numpy.sqrt(2*numpy.log(2)))
        FWO = 2.0*localf*numpy.tan(numpy.pi*self.noiseBWO/360.0)
        sigmaO = FWO/(2*numpy.sqrt(2*numpy.log(2)))
        yy, xx = numpy.mgrid[0:self._mysize, 0:self._mysize]
        xx = (0.5 - 1.0 / self._mysize * xx)
        yy = (0.5 - 1.0 / self._mysize * yy)
        filter=filters.make2DGauss(xx,yy,mean=(localf,0), sd=(sigmaF,sigmaO))
        filter=filter+filters.make2DGauss(xx,yy, mean=(-localf,0), sd=(sigmaF,sigmaO))
        filter = numpy.array(
                Image.fromarray(filter).rotate(
                        self.noiseOri,
                        Image.BICUBIC
                )
        )
        return FT*filter

    def updateNoise(self):
        """Updates the noise sample. Does not change any of the noise parameters 
            but choses a new random sample given the previously set parameters.
        """

        if not(self.noiseType in ['binary','Binary','normal','Normal','uniform','Uniform']):
            if (self.noiseType in ['image', 'Image']) and (self.imageComponent in ['amplitude','Amplitude']):
                self.noiseTex = numpy.random.uniform(0,1,int(self._mysize**2))
                self.noiseTex = numpy.reshape(self.noiseTex,(int(self._mysize),int(self._mysize)))
                if self.filter in ['Butterworth','butterworth']:
                    self.noiseTex = fftshift(self._filter(self.noiseTex))
                elif self.filter in ['Gabor','gabor']:
                    self.noiseTex = fftshift(self._gabor(self.noiseTex))
                elif self.filter in ['Isotropic','isotropic']:
                    self.noiseTex = fftshift(self._isotropic(self.noiseTex))
                self.noiseTex[0][0] = 0
                In = self.noiseTex * exp(1j*self.noisePh)
                Im = numpy.real(ifft2(In))
            else:
                Ph = numpy.random.uniform(0,2*numpy.pi,int(self._mysize**2))
                Ph = numpy.reshape(Ph,(int(self._mysize),int(self._mysize)))
                In = self.noiseTex * exp(1j*Ph)
                Im = numpy.real(ifft2(In))
                Im = ifftshift(Im)
            gsd = filters.getRMScontrast(Im)
            factor = gsd*self.noiseClip
            numpy.clip(Im, -factor, factor, Im)
            self.tex = Im / factor
        elif self.noiseType in ['normal','Normal']:
            self.noiseTex = numpy.random.randn(int(self._sideLength[1]),int(self._sideLength[0])) / self.noiseClip
        elif self.noiseType in ['uniform','Uniform']:
            self.noiseTex = 2.0 * numpy.random.rand(int(self._sideLength[1]),int(self._sideLength[0])) - 1.0
        else:
            numpy.random.shuffle(self.noiseTex)  # pick random noise sample by shuffleing values
            self.noiseTex = numpy.reshape(self.noiseTex,(int(self._sideLength[1]),int(self._sideLength[0])))
        if self.noiseType in ['binary','Binary','normal','Normal','uniform','Uniform']:
            if self.filter in ['butterworth', 'Butterworth', 'Gabor','gabor','Isotropic','isotropic']:
                if self.units == 'pix':
                    if self._mysize[0] == self._mysize[1]:
                        baseImage = numpy.array(
                                Image.fromarray(self.noiseTex).resize(
                                        (int(self._mysize[0]), int(self._mysize[1])),
                                        Image.NEAREST
                                )
                        )
                    else:
                        msg = ('NoiseStim can only apply filters to square noise images')
                        raise ValueError(msg)
                else:
                    baseImage = numpy.array(
                            Image.fromarray(self.noiseTex).resize(
                                    (int(self._mysize), int(self._mysize)),
                                    Image.NEAREST
                            )
                    )
                baseImage = numpy.array(baseImage).astype(
                        numpy.float32) * 0.0078431372549019607 - 1.0
                FT = fft2(baseImage)
                spectrum = numpy.absolute(fftshift(FT))
                angle = numpy.angle(FT)
                if self.filter in ['butterworth','Butterworth']:
                    spectrum = fftshift(self._filter(spectrum))
                elif self.filter in ['isotropic','Isotropic']:
                    spectrum = fftshift(self._isotropic(spectrum))
                elif self.filter in ['gabor','Gabor']:
                    spectrum = fftshift(self._gabor(spectrum))
                spectrum[0][0] = 0 # set DC to zero
                FT = spectrum * exp(1j*angle)
                
                Im = numpy.real(ifft2(FT))
                gsd = filters.getRMScontrast(Im)
                factor = gsd*self.noiseClip
                numpy.clip(Im, -factor, factor, Im)
                self.tex = Im / factor
            else:
                if not(self.noiseType in ['image','Image']):
                    self.tex = self.noiseTex
                
    
            
    def buildNoise(self):
        """build a new noise sample. Required to act on changes to any noise parameters or texRes.
        """

        if self.units == 'pix':
            if not (self.noiseType in ['Binary','binary','Normal','normal','uniform','Uniform']):
                mysize = numpy.max(self.size)
            else:
                mysize = self.size
            sampleSize = self.noiseElementSize
            mysf = self.__dict__['noiseBaseSf']*mysize
            lowsf = self.noiseFilterLower*numpy.max(mysize) # filter can only be applied to square images anyway
            upsf = self.noiseFilterUpper*numpy.max(mysize)
        else:
            mysize = self.texRes
            pixSize = self.size/self.texRes
            sampleSize = self.noiseElementSize/pixSize
            mysf = self.size[0]*self.noiseBaseSf
            lowsf = self.size[0]*self.noiseFilterLower
            upsf = self.size[0]*self.noiseFilterUpper
       
        self._mysize = mysize  # store for use by updateNoise()
        self._sf = mysf
        self._lowsf = lowsf
        self._upsf = upsf
        if self.noiseType in ['binary','Binary','normal','Normal','uniform','Uniform']:
            self._sideLength = numpy.round(mysize/sampleSize)  # dummy side length for use when unpacking noise samples in updateNoise()
            self._sideLength.astype(int)
            if ((self._sideLength[0] < 2) and (self._sideLength[1] < 2)):
                msg=('Noise sample size '
                     'must result in more than '
                     '1 sample per image dimension.')
                raise ValueError(msg)
            totalSamples = self._sideLength[0]*self._sideLength[1]
            if self.noiseType in ['binary','Binary']:
                self.noiseTex=numpy.append(numpy.ones(int(numpy.round(totalSamples/2.0))),-1*numpy.ones(int(numpy.round(totalSamples/2.0))))
        elif self.noiseType in ['White','white','filtered','Filtered','isotropic','Isotropic','Gabor','gabor']:
            self.noiseTex = numpy.ones((int(mysize),int(mysize)))
            self.noiseTex = fftshift(self.noiseTex)
        elif self.noiseType in ['Image','image']:
            if not(self.noiseImage in ['None','none']):  
                im = Image.open(self.noiseImage)
                im = im.transpose(Image.FLIP_TOP_BOTTOM)
                im = im.convert("L")  # FORCE TO LUMINANCE
                im = im.resize((int(self._mysize),int(self._mysize)),
                               Image.BILINEAR)
                intensity = numpy.array(im).astype(
                        numpy.float32) * 0.0078431372549019607 - 1.0
                if self.imageComponent in ['phase', 'Phase']:
                    self.noiseTex = numpy.absolute(fftshift(fft2(intensity))) # fftshift here is undone later
                elif self.imageComponent in ['amplitude', 'Amplitude']:
                    self.noisePh = numpy.angle((fft2(intensity))) # fftshift here is undone later
                    self.noiseTex = numpy.random.uniform(0,1,int(self._mysize**2))
                    self.noiseTex = numpy.reshape(self.noiseTex,(int(self._mysize),int(self._mysize)))
                else:
                    raise ValueError("Unknown value for imageComponent in noiseStim")
            else:
                self.noiseTex = numpy.ones((int(mysize),int(mysize)))  # if image is 'None' will make white noise as temporary measure
        else:
            msg = ('Noise type not recognised. Valid types are Binary, Uniform, Normal,'
                    'White, Filtered, Gabor, Isotropic or Image')
            raise ValueError(msg)
        if not(self.noiseType in ['binary','Binary','normal','Normal','uniform','Uniform']):
            if (self.noiseType in ['filtered','Filtered']) or (self.filter in ['butterworth', 'Butterworth']):
                self.noiseTex=self._filter(self.noiseTex)
            elif (self.noiseType in ['Isotropic','isotropic']) or (self.filter in ['isotropic', 'Isotropic']):
                self.noiseTex = self._isotropic(self.noiseTex)
            elif (self.noiseType in ['Gabor','gabor']) or (self.filter in ['gabor', 'Gabor']):
                self.noiseTex = self._gabor(self.noiseTex)
            self.noiseTex = fftshift(self.noiseTex)
            self.noiseTex[0][0] = 0 # Set DC to zero
  
        self._needBuild = False # prevent noise from being re-built at next draw() unless a parameter is changed in the mean time.
        self.updateNoise()  # now choose the initial random sample.



# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

_is_epydoc = False

# Pen digitizers /tablets that support Wintab API
from .. import Device, Computer
from ...constants import EventConstants, DeviceConstants
from ...errors import print2err, printExceptionDetailsToStdErr
import numpy as N
import copy

from psychopy import platform_specific
_sendStayAwake = platform_specific.sendStayAwake

class SimulatedWinTabPacket:
    _next_pkt_id = 1
    def __init__(self, time, x, y, press, buttons=0):
        self.pkTime=time*1000.0
        self.pkStatus=0
        self.pkSerialNumber=self.getNextID()
        self.pkButtons=buttons
        self.pkX = x                 
        self.pkY = y                 
        self.pkZ = 0                 
        self.pkNormalPressure=press          #('pressure',N.uint32),
        
        class Orientation:
            def __init__(self):
                self.orAzimuth=0         #('orient_azimuth',N.int32),
                self.orAltitude=0        #('orient_altitude;',N.int32),
                self.orTwist=0           #('orient_twist',N.int32),
                
        self.pkOrientation=Orientation()
        self.pkOrientation.orAzimuth=0         #('orient_azimuth',N.int32),
        self.pkOrientation.orAltitude=0        #('orient_altitude;',N.int32),
        self.pkOrientation.orTwist=0           #('orient_twist',N.int32),
    
    @classmethod    
    def getNextID(cls):
        v = cls._next_pkt_id
        cls._next_pkt_id+=1
        return v
    
class Wintab(Device):
    """The Wintab class docstr TBC."""
    EVENT_CLASS_NAMES = ['WintabSampleEvent',
                         'WintabEnterRegionEvent',
                         'WintabLeaveRegionEvent']

    DEVICE_TYPE_ID = DeviceConstants.WINTAB
    DEVICE_TYPE_STRING = 'WINTAB'

    __slots__ = ['_wtablets',
                 '_wtab_shadow_windows',
                 '_wtab_canvases',
                 '_last_sample',
                 '_first_hw_and_hub_times',
                 '_mouse_sim',
                 '_ioMouse',
                 '_simulated_wintab_events',
                 '_last_simulated_evt',
                 '_mouse_leave_region_timeout'
                 ]

    def __init__(self, *args, **kwargs):
        Device.__init__(self, *args, **kwargs['dconfig'])
        self._wtablets = []
        self._wtab_shadow_windows = []
        self._wtab_canvases = []
        
        self._ioMouse = self._last_simulated_evt = None
        self._mouse_sim = self.getConfiguration().get('mouse_simulation').get('enable')
        self._mouse_leave_region_timeout = self.getConfiguration().get('mouse_simulation').get('leave_region_timeout')
        self._simulated_wintab_events = []
        if self._mouse_sim:        
            self._registerMouseMonitor()
            self._setHardwareInterfaceStatus(True)
        else:
            self._init_wintab()

        # Following are used for sample status tracking
        self._last_sample = None
        self._first_hw_and_hub_times = None

    def _init_wintab(self):

        if Computer.platform == 'win32' and Computer.is_iohub_process:
            try:
                from .win32 import get_tablets
            except Exception:
                self._setHardwareInterfaceStatus(
                    False, u"Error: ioHub Wintab Device "
                    u"requires wintab32.dll to be "
                    u"installed.")

                def get_tablets():
                    return []

        else:
            def get_tablets():
                print2err('Error: iohub.devices.wintab only supports '
                          'Windows OS at this time.')
                return []

            self._setHardwareInterfaceStatus(False,
                                             u"Error:ioHub Wintab Device "
                                             u"only supports Windows OS "
                                             u"at this time.")

        self._wtablets = get_tablets()

        index = self.getConfiguration().get('device_number', 0)

        if self._wtablets is None:
            return False
        if len(self._wtablets) == 0:
            self._setHardwareInterfaceStatus(False,
                                             u"Error: No WinTab Devices"
                                             u" Detected.")
            return False
        if index >= len(self._wtablets):
            self._setHardwareInterfaceStatus(
                False, u"Error: device_number {} "
                u"is out of range. Only {} "
                u"WinTab devices detected.". format(
                    index, len(
                        self._wtablets)))
            return False

        exp_screen_info = self._display_device.getRuntimeInfo()
        swidth, sheight = exp_screen_info.get('pixel_resolution', [None, None])
        screen_index = exp_screen_info.get('index', 0)
        if swidth is None:
            self._setHardwareInterfaceStatus(False,
                                             u"Error: Wintab device is"
                                             u" unable to query experiment "
                                             u"screen pixel_resolution.")
            return False

        from pyglet.window import Window
        self._wtab_shadow_windows.append(
            Window(
                width=swidth,
                height=sheight,
                visible=False,
                fullscreen=True,
                vsync=False,
                screen=screen_index))
        self._wtab_shadow_windows[0].switch_to()
        self._wtab_shadow_windows[0].set_mouse_visible(False)

        from pyglet import app
        app.windows.remove(self._wtab_shadow_windows[0])

        try:
            self._wtab_canvases.append(
                self._wtablets[index].open(self._wtab_shadow_windows[0], self))
        except Exception as e:
            self._setHardwareInterfaceStatus(False,
                                             u"Error: Unable to create"
                                             u"WintabCanvas for device."
                                             u"Exception: {}".
                                             format(e))
            return False

        self._setHardwareInterfaceStatus(True)
        return True

    def getHardwareConfig(self, index=0):
        if self._ioMouse is None:
            hw_model_info = self._wtablets[index].hw_model_info
            return {'Context': self._wtab_canvases[index].getContextInfo(),
                    'Axis': self._wtablets[index].hw_axis_info,
                    'ModelInfo': hw_model_info
                    }

        swidth, sheight = self._display_device.getRuntimeInfo().get('pixel_resolution',[None, None])

        # running in mouse sim mode, so create hw config
        axis_info = {'orient_altitude': {'min': 0, 'max': 0,  'adjust': 0.0, 'factor': 0.0, 'units': 0,
                                         'resolution': 0},
                     'orient_azimuth': {'min': 0, 'max': 0, 'factor': 0.0, 'units': 0, 'resolution': 0}, 
                     'pressure': {'min': 0, 'max': 1, 'units': 0, 'resolution': 0}, 
                     'orient_twist': {'min': 0, 'max': 0, 'units': 0, 'resolution': 0}, 
                     'y': {'min': 0, 'max': sheight, 'units': 0, 'resolution': 0}, 
                     'x': {'min': 0, 'max': swidth, 'units': 0, 'resolution': 0}, 
                     'z': {'min': 0, 'max': 0, 'units': 0, 'resolution': 0}}
        context =  {'lcPktMode': 0, 'lcMoveMask': 0, 'lcLocks': 0, 'lcOptions': 0, 'lcInExtX': 0, 'lcSysOrgY': 0,
                    'lcSysOrgX': -0, 'lcPktRate': 0, 'lcBtnDnMask': 0, 'lcSysSensY': 0, 'lcSysSensX': 0,
                    'lcBtnUpMask': 0, 'lcSensY': 0, 'lcSensX': 0, 'lcSensZ': 0, 'lcInOrgX': 0, 'lcInOrgY': 0,
                    'lcInOrgZ': 0, 'lcOutExtX': 0, 'lcOutOrgX': 0, 'lcSysMode': 0, 'lcPktData': 0, 'lcOutOrgZ': 0,
                    'lcOutExtZ': 0, 'lcOutExtY': 0, 'lcOutOrgY': 0, 'lcInExtY': 0, 'lcStatus': 0, 'lcInExtZ': 0,
                    'lcName': '', 'lcDevice': 0, 'lcSysExtX': 0, 'lcMsgBase': 0, 'lcSysExtY': 0}
        model_info =  {'type': 0, 'handle': 0, 'name': 'Mouse Simulation Mode', 'id': ''}
                    
        return {"Context":context,
                 "Axis":axis_info,
                 "ModelInfo":model_info
                }

    def enableEventReporting(self, enabled=True):
        for wtc in self._wtab_canvases:
            wtc.enable(enabled)

        _sendStayAwake()
        
        if self.isReportingEvents() != enabled:
            self._simulated_wintab_events = []
            self._last_sample = None
            self._first_hw_and_hub_times = None
            self._last_simulated_evt = None

        return Device.enableEventReporting(self, enabled)

    def _addSimulatedWintabEvent(self,packet):            
        w, h = self._display_device.getRuntimeInfo().get('pixel_resolution')
        w = w/2
        h=h/2
        cevt = [EventConstants.WINTAB_SAMPLE,
                packet.pkTime/1000.0,
                packet.pkStatus,
                packet.pkSerialNumber,
                packet.pkButtons,
                packet.pkX+w,                 #('x',N.int32),
                packet.pkY+h,                 #('y',N.int32),
                packet.pkZ,                 #('z',N.int32),
                packet.pkNormalPressure,          #('pressure',N.uint32),
                packet.pkOrientation.orAzimuth,         #('orient_azimuth',N.int32),
                packet.pkOrientation.orAltitude,        #('orient_altitude;',N.int32),
                packet.pkOrientation.orTwist,           #('orient_twist',N.int32),
                0          #('status', N.uint8)
                ]                
        self._simulated_wintab_events.append(cevt)

    def _injectEventsIfNeeded(self, events):
        """
        """
        if len(events):
            if self._last_simulated_evt is None:
                events.insert(0, [EventConstants.WINTAB_ENTER_REGION,
                              0,
                              0,
                              0,
                              0,
                              0,                 #('x',N.int32),
                              0,                 #('y',N.int32),
                              0,                 #('z',N.int32),
                              0,          #('pressure',N.uint32),
                              0,         #('orient_azimuth',N.int32),
                              0,        #('orient_altitude;',N.int32),
                              0,           #('orient_twist',N.int32),
                              0])
            self._last_simulated_evt = events[-1]
            
        elif self._last_simulated_evt:
            epress = self._last_simulated_evt[8]            
            if not epress:
                ctime = Computer.getTime()
                etime = self._last_simulated_evt[1]  
                if ctime-etime > self._mouse_leave_region_timeout:
                    events.append([EventConstants.WINTAB_LEAVE_REGION,
                                        0,
                                        0,
                                        0,
                                        0,
                                        0,                 #('x',N.int32),
                                        0,                 #('y',N.int32),
                                        0,                 #('z',N.int32),
                                        0,          #('pressure',N.uint32),
                                        0,         #('orient_azimuth',N.int32),
                                        0,        #('orient_altitude;',N.int32),
                                        0,           #('orient_twist',N.int32),
                                        0])
                    self._last_simulated_evt = None
        return events
    
    def _poll(self):
        try:
            for swin in self._wtab_shadow_windows:
                swin.switch_to()
                swin.dispatch_events()
            logged_time = Computer.getTime()

            if not self.isReportingEvents():
                self._last_poll_time = logged_time
                self._simulated_wintab_events = []
                for wtc in self._wtab_canvases:
                    del wtc._iohub_events[:]
                return False

            confidence_interval = logged_time - self._last_poll_time
            # Using 0 delay for now as it is unknown.
            delay = 0.0

            wintab_events = []
            if self._ioMouse:
                wintab_events = self._simulated_wintab_events
                wintab_events = self._injectEventsIfNeeded(wintab_events)
                self._simulated_wintab_events = []
            else: 
                for wtc in self._wtab_canvases:
                    wintab_events.extend(copy.deepcopy(wtc._iohub_events))
                    del wtc._iohub_events[:]


            for wte in wintab_events:
                if wte and wte[0] != EventConstants.WINTAB_SAMPLE:
                    # event is enter or leave region type, so clear
                    # last sample as a flag that next sample should
                    # be FIRST_ENTER
                    self._last_sample = None
                else:
                    if self._first_hw_and_hub_times is None:
                        self._first_hw_and_hub_times = wte[1], logged_time

                    status = 0
                    cur_press_state = wte[8]
                    if self._last_sample is None:
                        status += WintabSampleEvent.STATES[
                            'FIRST_ENTER']
                        if cur_press_state > 0:
                            status += WintabSampleEvent.STATES[
                                'FIRST_PRESS']
                            status += WintabSampleEvent.STATES[
                                'PRESSED']
                        else:
                            status += WintabSampleEvent.STATES[
                                'FIRST_HOVER']
                            status += WintabSampleEvent.STATES[
                                'HOVERING']
                    else:
                        prev_press_state = self._last_sample[8]
                        if cur_press_state > 0:
                            status += WintabSampleEvent.STATES[
                                'PRESSED']
                        else:
                            status += WintabSampleEvent.STATES[
                                'HOVERING']

                        if cur_press_state > 0 and prev_press_state == 0:
                            status += WintabSampleEvent.STATES[
                                'FIRST_PRESS']
                        elif cur_press_state == 0 and prev_press_state > 0:
                            status += WintabSampleEvent.STATES[
                                'FIRST_HOVER']
                    # Fill in status field based on previous sample.......
                    wte[-1] = status
                    self._last_sample = wte

                self._addNativeEventToBuffer((logged_time,
                                              delay,
                                              confidence_interval,
                                              wte))

            self._last_poll_time = logged_time
            return True
        except Exception as e:
            print2err('ERROR in WintabDevice._poll: ', e)
            printExceptionDetailsToStdErr()

    def _getIOHubEventObject(self, native_event_data):
        '''

        :param native_event_data:
        :return:
        '''
        logged_time, delay, confidence_interval, wt_event = native_event_data
        evt_type = wt_event[0]
        device_time = wt_event[1]
        #evt_status = wt_event[2]

        # TODO: Correct for polling interval / CI when calculating iohub_time
        iohub_time = logged_time
        if self._first_hw_and_hub_times:
            hwtime, iotime = self._first_hw_and_hub_times
            iohub_time = iotime + (wt_event[1] - hwtime)

        ioevt = [0, 0, 0, Device._getNextEventID(),
                 evt_type,
                 device_time,
                 logged_time,
                 iohub_time,
                 confidence_interval,
                 delay,
                 0
                 ]
        ioevt.extend(wt_event[3:])
        return ioevt

    def _close(self):
        for wtc in self._wtab_canvases:
            wtc.close()
        for swin in self._wtab_shadow_windows:
            swin.close()
        if self._ioMouse:
            self._unregisterMouseMonitor()
            self._ioMouse=None
            self._last_simulated_evt=None
        Device._close(self)

    def _registerMouseMonitor(self):
        self._ioMouse=mouseDevice=None
        self._last_simulated_evt=None
        if self._iohub_server:
            for dev in self._iohub_server.devices:
                if dev.__class__.__name__ == 'Mouse':
                    mouseDevice=dev

        if mouseDevice:
            eventIDs=[EventConstants.MOUSE_BUTTON_PRESS,
                      EventConstants.MOUSE_BUTTON_RELEASE,
                      EventConstants.MOUSE_MOVE,
                      EventConstants.MOUSE_DRAG                      
                      ]
            self._ioMouse=mouseDevice
            self._ioMouse._addEventListener(self,eventIDs)
        else:
            print2err("Warning: elCG could not connect to Mouse device for events.")

    def _unregisterMouseMonitor(self):
        if self._ioMouse:
            self._ioMouse._removeEventListener(self)

    def _handleEvent(self, event):
        """
        """
        event_type_index = DeviceEvent.EVENT_TYPE_ID_INDEX
        tix = DeviceEvent.EVENT_HUB_TIME_INDEX
        bix = 14
        xix = 15
        yix = 16
        if event[event_type_index] == EventConstants.MOUSE_BUTTON_PRESS:
            if event[-10]==1:
                self._addSimulatedWintabEvent(SimulatedWinTabPacket(time=event[tix],
                                                                x=event[xix],
                                                                y=event[yix],
                                                                press=1.0,
                                                                buttons=event[bix]))                                                                
        elif event[event_type_index] == EventConstants.MOUSE_BUTTON_RELEASE:
            if event[-10]==1:
                self._addSimulatedWintabEvent(SimulatedWinTabPacket(time=event[tix],
                                                                x=event[xix],
                                                                y=event[yix],
                                                                press=0.0,
                                                                buttons=event[bix]))
        elif event[event_type_index] == EventConstants.MOUSE_MOVE:
            self._addSimulatedWintabEvent(SimulatedWinTabPacket(time=event[tix],
                                                                x=event[xix],
                                                                y=event[yix],
                                                                press=0.0,
                                                                buttons=event[bix]))
        elif event[event_type_index] == EventConstants.MOUSE_DRAG:
            if event[-10]==1 or event[bix]==1:
                self._addSimulatedWintabEvent(SimulatedWinTabPacket(time=event[tix],
                                                                    x=event[xix],
                                                                    y=event[yix],
                                                                    press=1.0,
                                                                    buttons=event[bix]))
        else:
            Device._handleEvent(self, event)

############# Wintab Event Classes ####################

from .. import DeviceEvent


class WintabInputEvent(DeviceEvent):
    """The WintabInputEvent is an abstract class that ......."""
    PARENT_DEVICE = Wintab
    _newDataTypes = []

    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):
        DeviceEvent.__init__(self, *args, **kwargs)


class WintabSampleEvent(WintabInputEvent):
    """WintabSampleEvent's occur when....."""
    EVENT_TYPE_STRING = 'WINTAB_SAMPLE'
    EVENT_TYPE_ID = EventConstants.WINTAB_SAMPLE
    IOHUB_DATA_TABLE = EVENT_TYPE_STRING

    STATES = dict()
    # A sample that is the first sample following a time gap in the sample
    # stream
    STATES['FIRST_ENTER'] = 1
    # A sample that is the first sample with pressure == 0
    # following a sample with pressure > 0
    STATES['FIRST_HOVER'] = 2
    # A sample that has pressure == 0, and previous sample also had pressure
    # == 0
    STATES['HOVERING'] = 4
    # A sample that is the first sample with pressure > 0
    # following a sample with pressure == 0
    STATES['FIRST_PRESS'] = 8
    #  A sample that has pressure > 0
    # following a sample with pressure > 0
    STATES['PRESSED'] = 16
	
    tstates = dict()
    for k, v in STATES.items():
        tstates[v] = k
    for k, v in tstates.items():
        STATES[k] = v

    _newDataTypes = [
        ('serial_number', N.uint32),
        ('buttons', N.int32),
        ('x', N.int32),
        ('y', N.int32),
        ('z', N.int32),
        ('pressure', N.uint32),
        ('orient_azimuth', N.int32),
        ('orient_altitude', N.int32),
        ('orient_twist', N.int32),
        ('status', N.uint8)
    ]

    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):
        #: serial_number Hardware assigned PACKET serial number
        self.serial_number = None

        #: TODO: buttons
        self.buttons = None

        #: x Horizontal position of stylus on tablet surface.
        self.x = None

        #: y Vertical position of stylus on tablet surface.
        self.y = None

        #: z Distance of stylus tip from tablet surface
        #: Supported on Wacom Intuos4; other device support unknown.
        #: Value will between 0 and max_val, where max_val is usually 1024.
        #: A value of 0 = tip touching surface, while
        #: max_val = tip height above surface before events stop being reported.
        self.z = None

        #: pressure: Pressure of stylus tip on tablet surface.
        self.pressure = None

        #: orient_azimuth
        self.orient_azimuth = None

        #: orient_altitude
        self.orient_altitude = None

        #: orient_twist
        self.orient_twist = None
        WintabInputEvent.__init__(self, *args, **kwargs)


class WintabEnterRegionEvent(WintabSampleEvent):
    """
    TODO: WintabEnterRegionEvent doc str
    """
    EVENT_TYPE_STRING = 'WINTAB_ENTER_REGION'
    EVENT_TYPE_ID = EventConstants.WINTAB_ENTER_REGION
    IOHUB_DATA_TABLE = WintabSampleEvent.EVENT_TYPE_STRING

    def __init__(self, *args, **kwargs):
        WintabSampleEvent.__init__(self, *args, **kwargs)


class WintabLeaveRegionEvent(WintabSampleEvent):
    """
    TODO: WintabLeaveRegionEvent doc str
    """
    EVENT_TYPE_STRING = 'WINTAB_LEAVE_REGION'
    EVENT_TYPE_ID = EventConstants.WINTAB_LEAVE_REGION
    IOHUB_DATA_TABLE = WintabSampleEvent.EVENT_TYPE_STRING

    def __init__(self, *args, **kwargs):
        WintabSampleEvent.__init__(self, *args, **kwargs)



from .. import BaseStandaloneRoutine
from psychopy.localization import _translate
from psychopy.experiment import Param
from pathlib import Path
from psychopy.alerts import alert


class EyetrackerCalibrationRoutine(BaseStandaloneRoutine):
    categories = ['Eyetracking']
    targets = ["PsychoPy"]
    iconFile = Path(__file__).parent / "eyetracker_calib.png"
    tooltip = _translate("Calibration routine for eyetrackers")
    beta = True

    def __init__(self, exp, name='calibration',
                 progressMode="time", targetDur=1.5, expandDur=1, expandScale=1.5,
                 movementAnimation=True, movementDur=1.0, targetDelay=1.0,
                 innerFillColor='green', innerBorderColor='black', innerBorderWidth=2, innerRadius=0.0035,
                 fillColor='', borderColor="black", borderWidth=2, outerRadius=0.01,
                 colorSpace="rgb", units='from exp settings',
                 targetLayout="NINE_POINTS", randomisePos=True, textColor='white',
                 disabled=False
                 ):
        # Initialise base routine
        BaseStandaloneRoutine.__init__(self, exp, name=name, disabled=disabled)

        self.exp.requirePsychopyLibs(['iohub', 'hardware'])

        # Basic params
        self.order += [
            "targetLayout",
            "randomisePos",
            "textColor"
        ]

        del self.params['stopVal']
        del self.params['stopType']

        self.params['targetLayout'] = Param(targetLayout,
                                            valType='str', inputType="choice", categ='Basic',
                                            allowedVals=['THREE_POINTS', 'FIVE_POINTS', 'NINE_POINTS', "THIRTEEN_POINTS"],
                                            hint=_translate("Pre-defined target layouts"),
                                            label=_translate("Target Layout"))

        self.params['randomisePos'] = Param(randomisePos,
                                            valType='bool', inputType="bool", categ='Basic',
                                            hint=_translate("Should the order of target positions be randomised?"),
                                            label=_translate("Randomise Target Positions"))
        self.params['textColor'] = Param(textColor,
                                     valType='color', inputType="color", categ='Basic',
                                     hint=_translate("Text foreground color"),
                                     label=_translate("Text Color"))
        # Target Params
        self.order += [
            "targetStyle",
            "fillColor",
            "borderColor",
            "innerFillColor",
            "innerBorderColor",
            "colorSpace",
            "borderWidth",
            "innerBorderWidth",
            "outerRadius",
            "innerRadius",
        ]

        self.params['innerFillColor'] = Param(innerFillColor,
                                     valType='color', inputType="color", categ='Target',
                                     hint=_translate("Fill color of the inner part of the target"),
                                     label=_translate("Inner Fill Color"))

        self.params['innerBorderColor'] = Param(innerBorderColor,
                                           valType='color', inputType="color", categ='Target',
                                           hint=_translate("Border color of the inner part of the target"),
                                           label=_translate("Inner Border Color"))

        self.params['fillColor'] = Param(fillColor,
                                         valType='color', inputType="color", categ='Target',
                                         hint=_translate("Fill color of the outer part of the target"),
                                         label=_translate("Outer Fill Color"))

        self.params['borderColor'] = Param(borderColor,
                                           valType='color', inputType="color", categ='Target',
                                           hint=_translate("Border color of the outer part of the target"),
                                           label=_translate("Outer Border Color"))

        self.params['colorSpace'] = Param(colorSpace,
                                          valType='str', inputType="choice", categ='Target',
                                          allowedVals=['rgb', 'dkl', 'lms', 'hsv'],
                                          hint=_translate(
                                              "In what format (color space) have you specified the colors? (rgb, dkl, lms, hsv)"),
                                          label=_translate("Color Space"))

        self.params['borderWidth'] = Param(borderWidth,
                                           valType='num', inputType="single", categ='Target',
                                           hint=_translate("Width of the line around the outer part of the target"),
                                           label=_translate("Outer Border Width"))

        self.params['innerBorderWidth'] = Param(innerBorderWidth,
                                           valType='num', inputType="single", categ='Target',
                                           hint=_translate("Width of the line around the inner part of the target"),
                                           label=_translate("Inner Border Width"))

        self.params['outerRadius'] = Param(outerRadius,
                                           valType='num', inputType="single", categ='Target',
                                           hint=_translate("Size (radius) of the outer part of the target"),
                                           label=_translate("Outer Radius"))

        self.params['innerRadius'] = Param(innerRadius,
                                           valType='num', inputType="single", categ='Target',
                                           hint=_translate("Size (radius) of the inner part of the target"),
                                           label=_translate("Inner Radius"))

        self.params['units'] = Param(units,
                                     valType='str', inputType="choice", categ='Target',
                                     allowedVals=['from exp settings'], direct=False,
                                     hint=_translate("Units of dimensions for this stimulus"),
                                     label=_translate("Spatial Units"))

        # Animation Params
        self.order += [
            "progressMode",
            "targetDur",
            "expandDur",
            "expandScale",
            "movementAnimation",
            "movementDur",
            "targetDelay"
        ]

        self.params['progressMode'] = Param(progressMode,
                                            valType="str", inputType="choice", categ="Animation",
                                            allowedVals=["space key", "time"],
                                            hint=_translate("Should the target move to the next position after a "
                                                            "keypress or after an amount of time?"),
                                            label=_translate("Progress Mode"))

        self.depends.append(
            {"dependsOn": "progressMode",  # must be param name
             "condition": "in ['time', 'either']",  # val to check for
             "param": "targetDur",  # param property to alter
             "true": "show",  # what to do with param if condition is True
             "false": "hide",  # permitted: hide, show, enable, disable
             }
        )

        self.params['targetDur'] = Param(targetDur,
                                         valType='num', inputType="single", categ='Animation',
                                         hint=_translate(
                                             "Time limit (s) after which progress to next position"),
                                         label=_translate("Target Duration"))

        self.depends.append(
            {"dependsOn": "progressMode",  # must be param name
             "condition": "in ['space key', 'either']",  # val to check for
             "param": "expandDur",  # param property to alter
             "true": "show",  # what to do with param if condition is True
             "false": "hide",  # permitted: hide, show, enable, disable
             }
        )

        self.params['expandDur'] = Param(expandDur,
                                         valType='num', inputType="single", categ='Animation',
                                         hint=_translate(
                                             "Duration of the target expand/contract animation"),
                                         label=_translate("Expand / Contract Duration"))

        self.params['expandScale'] = Param(expandScale,
                                           valType='num', inputType="single", categ='Animation',
                                           hint=_translate("How many times bigger than its size the target grows"),
                                           label=_translate("Expand Scale"))

        self.params['movementAnimation'] = Param(movementAnimation,
                                                 valType='bool', inputType="bool", categ='Animation',
                                                 hint=_translate(
                                                     "Enable / disable animations as target stim changes position"),
                                                 label=_translate("Animate Position Changes"))

        self.depends.append(
            {"dependsOn": "movementAnimation",  # must be param name
             "condition": "== True",  # val to check for
             "param": "movementDur",  # param property to alter
             "true": "show",  # what to do with param if condition is True
             "false": "hide",  # permitted: hide, show, enable, disable
             }
        )

        self.params['movementDur'] = Param(movementDur,
                                           valType='num', inputType="single", categ='Animation',
                                           hint=_translate(
                                               "Duration of the animation during position changes."),
                                           label=_translate("Movement Duration"))

        self.depends.append(
            {"dependsOn": "movementAnimation",  # must be param name
             "condition": "== False",  # val to check for
             "param": "targetDelay",  # param property to alter
             "true": "show",  # what to do with param if condition is True
             "false": "hide",  # permitted: hide, show, enable, disable
             }
        )

        self.params['targetDelay'] = Param(targetDelay,
                                           valType='num', inputType="single", categ='Animation',
                                           hint=_translate(
                                               "Duration of the delay between positions."),
                                           label=_translate("Target Delay"))

    def writeMainCode(self, buff):
        # Alert user if eyetracking isn't setup
        if self.exp.eyetracking == "None":
            alert(code=4505)
        # Get inits
        inits = self.params
        # Code-ify 'from exp settings'
        if self.params['units'].val == 'from exp settings':
            inits['units'].val = None
        # Synonymise expand dur and target dur
        if inits['progressMode'].val == 'time':
            inits['expandDur'] = inits['targetDur']
        if inits['progressMode'].val == 'space key':
            inits['targetDur'] = inits['expandDur']
        # Synonymise movement dur and target delay
        if inits['movementAnimation'].val:
            inits['targetDelay'] = inits['movementDur']
        else:
            inits['movementDur'] = inits['targetDelay']

        BaseStandaloneRoutine.writeMainCode(self, buff)

        # Make target
        code = (
            "# define target for %(name)s\n"
            "%(name)sTarget = visual.TargetStim(win, \n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                "name='%(name)sTarget',\n"
                "radius=%(outerRadius)s, fillColor=%(fillColor)s, borderColor=%(borderColor)s, lineWidth=%(borderWidth)s,\n"
                "innerRadius=%(innerRadius)s, innerFillColor=%(innerFillColor)s, innerBorderColor=%(innerBorderColor)s, innerLineWidth=%(innerBorderWidth)s,\n"
                "colorSpace=%(colorSpace)s, units=%(units)s\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
            ")"
        )
        buff.writeIndentedLines(code % inits)
        # Make config object
        code = (
            "# define parameters for %(name)s\n"
            "%(name)s = hardware.eyetracker.EyetrackerCalibration(win, \n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(1, relative=True)
        code = (
                "eyetracker, %(name)sTarget,\n"
                "units=%(units)s, colorSpace=%(colorSpace)s,\n"
                "progressMode=%(progressMode)s, targetDur=%(targetDur)s, expandScale=%(expandScale)s,\n"
                "targetLayout=%(targetLayout)s, randomisePos=%(randomisePos)s, textColor=%(textColor)s,\n"
                "movementAnimation=%(movementAnimation)s, targetDelay=%(targetDelay)s\n"
        )
        buff.writeIndentedLines(code % inits)
        buff.setIndentLevel(-1, relative=True)
        code = (
            ")\n"
            "# run calibration\n"
            "%(name)s.run()\n"
            "# clear any keypresses from during %(name)s so they don't interfere with the experiment\n"
            "defaultKeyboard.clearEvents()\n"
        )
        buff.writeIndentedLines(code % inits)



# -*- coding: utf-8 -*-
# Part of the PsychoPy library
# Copyright (C) 2012-2020 iSolver Software Solutions (C) 2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).
"""
Eye Tracker Validation procedure using the ioHub common eye tracker interface.

To use the validation process from within a Coder script:
* Create a target stim
* Create a list of validation target positions. Use the `PositionGrid` class to help create a target position list.
* Create a ValidationProcedure class instance, providing the target stim and position list and other arguments
  to define details of the validation procedure.
* Use `ValidationProcedure.run()` to perform the validation routine.
* Use `ValidationProcedure.getValidationResults()` to access information about each target position displayed and
  the events collected during the each target validation period.

See demos/coder/iohub/eyetracking/validation.py for a complete example.
"""
from weakref import proxy
import numpy as np
from time import sleep
import os
import sys
from matplotlib import pyplot as pl

from psychopy import visual
from psychopy.iohub.util import win32MessagePump, normjoin
from psychopy.iohub.constants import EventConstants
from psychopy.iohub.client import ioHubConnection, Computer
from psychopy.tools.monitorunittools import convertToPix
from psychopy.tools.monitorunittools import pix2deg, deg2pix

from psychopy.iohub.client.eyetracker.validation import PositionGrid, Trigger, KeyboardTrigger, TimeTrigger

getTime = Computer.getTime


class TargetStim:
    def __init__(self, win, radius=None, fillcolor=None, edgecolor=None, edgewidth=None,
                 dotcolor=None, dotradius=None, units=None, colorspace=None, opacity=1.0, contrast=1.0):
        """
        TargetStim is a 'doughnut' style target graphic used during the validation procedure.

        :param win: Window being used for validation.
        :param radius: The outer radius of the target.
        :param fillcolor: The color used to fill the target body.
        :param edgecolor: The color for the edge around the target.
        :param edgewidth: The thickness of the target outer edge (always in pixels).
        :param dotcolor: The color of the central target dot.
        :param dotradius: The radius to use for the target dot.
        :param units: The psychopy unit type of any size values.
        :param colorspace: The psychopy color space of any colors.
        :param opacity: The transparency of the target (0.0 - 1.0).
        :param contrast: The contrast of the target stim.
        """
        from weakref import proxy
        self.win = proxy(win)
        self.stim = []
        self._radius = radius
        outer = visual.Circle(self.win, radius=radius, fillColor=fillcolor, lineColor=edgecolor, lineWidth=edgewidth,
                              edges=32, units=units, colorSpace=colorspace, opacity=opacity,
                              contrast=contrast, interpolate=True, autoLog=False)
        self.stim.append(outer)

        if dotcolor and dotcolor != fillcolor:
            centerdot = visual.Circle(self.win, radius=dotradius, fillColor=dotcolor, lineColor=dotcolor,
                                      lineWidth=0.0, edges=32, interpolate=True, units=units,
                                      colorSpace=colorspace, opacity=opacity, contrast=contrast, autoLog=False)
            self.stim.append(centerdot)

    def setPos(self, pos):
        """
        Set the center position of the target stim. Used during validation procedure to
        change target position.
        """
        for s in self.stim:
            s.setPos(pos)

    @property
    def pos(self):
        return self.stim[0].pos

    @pos.setter
    def pos(self, value):
        self.setPos(value)

    @property
    def radius(self):
        return self._radius

    @radius.setter
    def radius(self, r):
        self._radius = self.stim[0].radius = r

    def setSize(self, s):
        """
        Update the size of the target stim.
        """
        self.stim[0].radius = s/2

    def getSize(self):
        """
        Get the size of the target stim.
        """
        return self.stim[0].radius*2

    def draw(self):
        """
        Draw the Target stim.
        """
        for s in self.stim:
            s.draw()

    def contains(self, p):
        """
        Is point p contained within the Target Stim?
        :param p: x, y position in stim units
        :return: bool: True: p is within the stim
        """
        return self.stim[0].contains(p)

    @property
    def innerRadius(self):
        try:
            return self.stim[1].radius
        except:
            return self.stim[0].radius

def create3PointGrid():
    io = ioHubConnection.getActiveConnection()
    if io is None:
        raise RuntimeError("iohub must be running.")
    l, t, r, b = io.devices.display.getCoordBounds()
    return [(0.0, (t-b)/4), (-(r-l)/4, -(t-b)/4), ((r-l)/4, -(t-b)/4)]


def create5PointGrid():
    io = ioHubConnection.getActiveConnection()
    if io is None:
        raise RuntimeError("iohub must be running.")
    four_point = PositionGrid(io.devices.display.getCoordBounds(), (2, 2), scale=0.85).getPositions()
    return [(0.0, 0.0),] + four_point


def create9PointGrid():
    io = ioHubConnection.getActiveConnection()
    if io is None:
        raise RuntimeError("iohub must be running.")
    return PositionGrid(io.devices.display.getCoordBounds(), (3, 3), scale=0.85, firstposindex=4)


def create13PointGrid():
    io = ioHubConnection.getActiveConnection()
    if io is None:
        raise RuntimeError("iohub must be running.")
    nine_point = create9PointGrid().getPositions()
    four_point = PositionGrid(io.devices.display.getCoordBounds(), (2, 2), scale=0.5).getPositions()
    thirteen_point = nine_point + four_point
    return thirteen_point


def create17PointGrid():
    io = ioHubConnection.getActiveConnection()
    if io is None:
        raise RuntimeError("iohub must be running.")
    sixteen_pos = PositionGrid(io.devices.display.getCoordBounds(), (4, 4), scale=0.85).getPositions()
    return [(0.0, 0.0), ] + sixteen_pos


class ValidationProcedure:
    def __init__(self,
                 win=None,  # psychopy window
                 target=None,  # target stim
                 positions=None,  # string constant or list of points
                 randomize_positions=True,  # boolean
                 expand_scale=None,  # float
                 target_duration=None,  # float
                 target_delay=None,  # float
                 enable_position_animation=True,
                 color_space=None,  # None == use window color space
                 unit_type=None,  # None == use window unit type (may need to enforce this for Validation)
                 progress_on_key=" ",  # str or None
                 gaze_cursor=None,  # None, color, or a stim object with setPos()
                 text_color=None,
                 show_results_screen=True,  # bool
                 save_results_screen=False,  # bool
                 # args not used by Builder at this time
                 contract_target=True,
                 accuracy_period_start=0.550,
                 accuracy_period_stop=.150,
                 show_intro_screen=False,
                 intro_text='Ready to Start Validation Procedure.',
                 results_in_degrees=False,
                 terminate_key="escape",
                 toggle_gaze_cursor_key="g"):
        """
        ValidationProcedure is used to test the gaze accuracy of a calibrated eye tracking system.

        Once a ValidationProcedure class instance has been created, call the `.run()` method
        to start the validation process, which consists of the following steps:

        1) (Optionally) Display an Introduction screen. A 'space' key press is used to start target presentation.
        2) Displaying the validation target at each position being validated. Target progression from one
           position to the next is controlled by the specified `triggers`, defaulting to a 'space' key press.
           The target graphics can simply jump from one position to the next, or optional target_animation settings
           can be used to have the target move across the screen from one point to the next and / or expand / contract
           at each target location.
        3) (Optionally) Display a Results screen. The Results screen shows each target position, the position of
           each sample used for the accuracy calculation, and some validation result statistics.

        Data collected during the validation target sequence is used to calculate accuracy information
        for each target position presented. The raw data as well as the computed accuracy stats is
        available via the `.getValidationResults()` method.

        To make the validation output consistent across iohub common eye tracker implementations, validation is
        performed on monocular eye data. If binocular eye samples are being recorded, the average of the
        left and right eye positions is used for each gaze sample.

        See the validation.py demo in demos.coder.iohub.eyetracking for a demo.

        TODO: Update param list docs
        :param win: PsychoPy window to use for validation. Must be full screen.
        :param target: Stimulus to use as validation target. If None, default `TargetStim` is used.
        :param positions: Positions to validate. Provide list of x,y pairs, or use a `PositionGrid` class.
        :param target_animation:
        :param randomize_positions: bool: Randomize target positions before presentation.
        :param show_intro_screen: bool: Display a validation procedure Introduction screen.
        :param intro_text: Introduction screen text.
        :param show_results_screen: bool: Display a validation procedure Results screen.
        :param results_in_degrees: bool: Convert results to visual degrees.
        :param save_results_screen: bool: Save results screen as image.
        :param terminate_key: Key that will end the validation procedure. Default is 'escape'.
        :param toggle_gaze_cursor_key: Key to toggle gaze cursor visibility (hidden to start). Default is key is 'g'.
        :param accuracy_period_start: Time prior to target trigger to use as start of period for valid samples.
        :param accuracy_period_stop: Time prior to target trigger to use as end of period for valid samples.
        :param triggers: Target progression triggers. Default is 'space' key press.
        :param storeeventsfor: iohub devices that events should be stored for.
        """
        self.terminate_key = terminate_key
        self.toggle_gaze_cursor_key = toggle_gaze_cursor_key

        self.io = ioHubConnection.getActiveConnection()

        if isinstance(positions, str):
            # position set constant, THREE_POINTS, FIVE_POINTS, NINE_POINTS, THIRTEEN_POINTS, SEVENTEEN_POINTS
            if positions == 'THREE_POINTS':
                positions = create3PointGrid()
            elif positions == 'FIVE_POINTS':
                positions = create5PointGrid()
            elif positions == 'NINE_POINTS':
                positions = create9PointGrid()
            elif positions == 'THIRTEEN_POINTS':
                positions = create13PointGrid()
            elif positions == 'SEVENTEEN_POINTS':
                positions = create17PointGrid()
            else:
                raise ValueError("Unsupported positions string constant: [{}]".format(positions))
        if isinstance(positions, (list, tuple)):
            positions = PositionGrid(posList=positions, firstposindex=0, repeatFirstPos=False)
        self.positions = positions

        self.randomize_positions = randomize_positions
        if self.randomize_positions:
            self.positions.randomize()

        self.win = proxy(win)

        target_animation = {}
        target_animation['enable'] = enable_position_animation
        target_animation['targetdelay'] = target_delay
        target_animation['targetduration'] = target_duration
        target_animation['expandedscale'] = expand_scale
        target_animation['contracttarget'] = contract_target

        self.animation_params = target_animation
        self.accuracy_period_start = accuracy_period_start

        self.accuracy_period_stop = accuracy_period_stop
        self.show_intro_screen = show_intro_screen
        self.intro_text = intro_text
        self.intro_text_stim = None
        self.show_results_screen = show_results_screen
        self.results_in_degrees = results_in_degrees
        self.save_results_screen = save_results_screen
        self._validation_results = None

        self.text_color = text_color
        self.text_color_space = color_space

        if text_color is None or text_color == 'auto':
            # If no calibration text color provided, base it on the window background color
            from psychopy.iohub.util import complement
            sbcolor = win.color
            from psychopy.colors import Color
            tcolor_obj = Color(sbcolor, win.colorSpace)
            self.text_color = complement(*tcolor_obj.rgb255)
            self.text_color_space = 'rgb255'

        storeeventsfor = [self.io.devices.keyboard,
                          self.io.devices.tracker,
                          self.io.devices.experiment]

        trig_list = []
        if progress_on_key:
            if isinstance(progress_on_key, (list, tuple)):
                for k in progress_on_key:
                    trig_list.append(KeyboardTrigger(k, on_press=True))
            else:
                trig_list.append(KeyboardTrigger(progress_on_key, on_press=True))
        elif target_duration:
            trig_list.append(TimeTrigger(start_time=None, delay=target_duration),)

        triggers = Trigger.getTriggersFrom(trig_list)

        # Create the ValidationTargetRenderer instance; used to control the sequential
        # presentation of the target at each of the grid positions.
        self.targetsequence = ValidationTargetRenderer(win, target=target, positions=self.positions,
                                                       triggers=triggers, storeeventsfor=storeeventsfor,
                                                       terminate_key=terminate_key,
                                                       gaze_cursor_key=toggle_gaze_cursor_key,
                                                       gaze_cursor=gaze_cursor,
                                                       color_space=color_space, unit_type=unit_type)

    def run(self):
        """
        Run the validation procedure, returning after the full validation process is complete, including:
            a) display of an instruction screen
            b) display of the target position sequence used for validation data collection.
            c) display of a validation accuracy results plot.
        """
        keyboard = self.io.devices.keyboard
        if self.show_intro_screen:
            # Display Validation Intro Screen
            self.showIntroScreen()
            if self.terminate_key and self.terminate_key in keyboard.waitForReleases(keys=[' ', 'space',self.terminate_key]):
                print("Escape key pressed. Exiting validation")
                self._validation_results = None
                return

        # Perform Validation.....
        terminate = not self.targetsequence.display(**self.animation_params)
        if terminate:
            print("Escape key pressed. Exiting validation")
            self._validation_results = None
            return

        self.io.clearEvents('all')

        self._createValidationResults()

        if self.show_results_screen:
            self.showResultsScreen()
            kb_presses = keyboard.waitForPresses(keys=['space',' ', self.terminate_key, self.targetsequence.gaze_cursor_key])
            while 'space' not in kb_presses and ' ' not in kb_presses:
                if self.targetsequence.gaze_cursor_key in kb_presses:
                    self.targetsequence.display_gaze = not self.targetsequence.display_gaze
                    self.showResultsScreen()
                if self.terminate_key in kb_presses:
                    print("Escape key pressed. Exiting validation")
                    break
                kb_presses = keyboard.waitForPresses(keys=['space', ' ',
                                                           self.terminate_key,
                                                           self.targetsequence.gaze_cursor_key])

        return self._validation_results

    def showResultsScreen(self):
        self.drawResultScreen()
        ftime = self.win.flip()
        if self.save_results_screen:
            self.win.getMovieFrame()
            self.win.saveMovieFrames(self._generateImageName())
        return ftime

    def showIntroScreen(self):
        text = self.intro_text + '\nPress SPACE to Start....'
        textpos = (0, 0)
        if self.intro_text_stim:
            self.intro_text_stim.setText(text)
            self.intro_text_stim.setPos(textpos)
        else:
            self.intro_text_stim = visual.TextStim(self.win, text=text, pos=textpos, height=30, color=self.text_color,
                                                   colorSpace=self.text_color_space, opacity=1.0, contrast=1.0, units='pix',
                                                   ori=0.0, antialias=True, bold=False, italic=False,
                                                   anchorHoriz='center', anchorVert='center',
                                                   wrapWidth=self.win.size[0] * .8)

        self.intro_text_stim.draw()
        self.win.flip()
        return self.win.flip()

    @property
    def results(self):
        """
        See getValidationResults().
        :return:
        """
        return self._validation_results

    def getValidationResults(self):
        """
        Return the validation results dict for the last validation run. If a validation as not yet been run(),
        None is returned. Validation results are provided separately for each target position and include:

        a) An array of the samples used for the accuracy calculation. The samples used are selected
           using the following criteria:
                i) Only samples where the target was stationary and not expanding or contracting are selected.
                ii) Samples are selected that fall between:

                              start_time_filter = last_sample_time - accuracy_period_start

                           and

                              end_time_filter = last_sample_time - accuracy_period_end

                           Therefore, the duration of the selected sample period is:

                              selection_period_dur = end_time_filter - start_time_filter

                iii) Sample that contain missing / invalid position data are removed, providing the
                     final set of samples used for accuracy calculations. The min, max, and mean values
                     from each set of selected samples is calculated.

        b) The x and y error of sampled gaze position relative to the current target position.
           This data is in the same units as is used by the validation window.

        c) The xy distance error from the from each eye's gaze position to the target position.
           This is also calculated as an average of both eyes when binocular data is available.
           The data is unsigned, providing the absolute distance from gaze to target positions

        Validation Results Dict Structure
        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

        {'display_bounds': [-1.0, 1.0, 1.0, -1.0],
         'display_pix': array([1920, 1080]),
         'display_units': 'norm',
         'max_error': 2.3668638421479,
         'mean_error': 0.9012516727129639,
         'min_error': 0.0,
         'passed': True,
         'position_count': 9,
         'positions_failed_processing': 0,
         'reporting_unit_type': 'degree',
         'target_positions': [array([0., 0.]), array([0.85, 0.85]), array([-0.85,  0.  ]),
                              array([0.85, 0.  ]), array([ 0.85, -0.85]), array([-0.85,  0.85]),
                              array([-0.85, -0.85]), array([0.  , 0.85]), array([ 0.  , -0.85])],
         'position_results': [{'index': 0,
                               'calculation_status': 'PASSED',
                               'target_position': array([0., 0.]),
                               'sample_time_range': [4.774341499977744, 6.8343414999777],
                               'filter_samples_time_range': [6.2843414999777005, 6.6843414999777],
                               'min_error': 0.0,
                               'max_error': 0.7484680652684592,
                               'mean_error': 0.39518431321527914,
                               'stdev_error': 0.24438398690651483,
                               'valid_filtered_sample_perc': 1.0,
                              },
                              # Validation results dict is given for each target position
                              # ....
                             ]
        }

        :return: validation results dict.
        """
        return self._validation_results

    def _createValidationResults(self):
        """
        Create validation results dict and save validation analysis info as experiment messages to
        the iohub .hdf5 file.

        :return: dict
        """
        self._validation_results = None
        sample_array = self.targetsequence.getSampleMessageData()
        target_positions_used = self.targetsequence.positions.getPositions()

        if self.results_in_degrees:
            for postdat in sample_array:
                postdat['targ_pos_x'], postdat['targ_pos_y'] = toDeg(self.win,
                                                                     *toPix(self.win, postdat['targ_pos_x'],
                                                                            postdat['targ_pos_y']))

                binoc_sample_types = [EventConstants.BINOCULAR_EYE_SAMPLE, EventConstants.GAZEPOINT_SAMPLE]
                if self.targetsequence.sample_type in binoc_sample_types:
                    postdat['left_eye_x'], postdat['left_eye_y'] = toDeg(self.win,
                                                                         *toPix(self.win, postdat['left_eye_x'],
                                                                                postdat['left_eye_y']))

                    postdat['right_eye_x'], postdat['right_eye_y'] = toDeg(self.win,
                                                                           *toPix(self.win, postdat['right_eye_x'],
                                                                                  postdat['right_eye_y']))
                else:
                    postdat['eye_x'], postdat['eye_y'] = toDeg(self.win,
                                                               *toPix(self.win, postdat['eye_x'], postdat['eye_y']))

        min_error = 100000.0
        max_error = 0.0
        summed_error = 0.0
        point_count = 0

        self.io.sendMessageEvent('Results', 'VALIDATION')
        results = dict(display_units=self.win.units, display_bounds=self.positions.bounds,
                       display_pix=self.win.size, position_count=len(sample_array),
                       target_positions=target_positions_used)

        for k, v in results.items():
            self.io.sendMessageEvent('{}: {}'.format(k, v), 'VALIDATION')

        results['position_results'] = []
        results['positions_failed_processing'] = 0

        for pindex, samplesforpos in enumerate(sample_array):
            self.io.sendMessageEvent('Target Position Results: {0}'.format(pindex), 'VALIDATION')

            stationary_samples = samplesforpos[samplesforpos['targ_state'] == self.targetsequence.TARGET_STATIONARY]

            if len(stationary_samples):
                last_stime = stationary_samples[-1]['eye_time']
                first_stime = stationary_samples[0]['eye_time']

                filter_stime = last_stime - self.accuracy_period_start
                filter_etime = last_stime - self.accuracy_period_stop

                all_samples_in_period = stationary_samples[stationary_samples['eye_time'] >= filter_stime]
                all_samples_in_period = all_samples_in_period[all_samples_in_period['eye_time'] < filter_etime]

                good_samples_in_period = all_samples_in_period[all_samples_in_period['eye_status'] == 0]

                all_samples_count = all_samples_in_period.shape[0]
                good_sample_count = good_samples_in_period.shape[0]
                try:
                    good_sample_ratio = good_sample_count / float(all_samples_count)
                except ZeroDivisionError:
                    good_sample_ratio = 0
            else:
                all_samples_in_period = []
                good_samples_in_period = []
                good_sample_ratio = 0

            # Dictionary of the different levels of samples selected during filtering
            # for valid samples to use in accuracy calculations.
            sample_msg_data_filtering = dict(all_samples=samplesforpos,  # All samples from target period.
                                                    # Sample during stationary period at end of target
                                                    # presentation display.
                                                    stationary_samples=stationary_samples,
                                                    # Samples that occurred within the
                                                    # defined time selection period.
                                                    time_filtered_samples=all_samples_in_period,
                                                    # Samples from the selection period that
                                                    # do not have missing data
                                                    used_samples=good_samples_in_period)

            position_results = dict(index=pindex,
                                    target_position=target_positions_used[pindex],
                                    sample_time_range=[first_stime, last_stime],
                                    filter_samples_time_range=[filter_stime, filter_etime],
                                    valid_filtered_sample_perc=good_sample_ratio)

            for k, v in position_results.items():
                self.io.sendMessageEvent('{}: {}'.format(k, v), 'VALIDATION')

            position_results['sample_from_filter_stages'] = sample_msg_data_filtering

            position_results2 = dict()
            if int(good_sample_ratio * 100) == 0:
                position_results2['calculation_status'] = 'FAILED'
                results['positions_failed_processing'] += 1
            else:
                target_x = good_samples_in_period[:]['targ_pos_x']
                target_y = good_samples_in_period[:]['targ_pos_y']

                binoc_sample_types = [EventConstants.BINOCULAR_EYE_SAMPLE, EventConstants.GAZEPOINT_SAMPLE]
                if self.targetsequence.sample_type in binoc_sample_types:
                    left_x = good_samples_in_period[:]['left_eye_x']
                    left_y = good_samples_in_period[:]['left_eye_y']
                    left_error_x = target_x - left_x
                    left_error_y = target_y - left_y
                    left_error_xy = np.hypot(left_error_x, left_error_y)

                    right_x = good_samples_in_period[:]['right_eye_x']
                    right_y = good_samples_in_period[:]['right_eye_y']
                    right_error_x = target_x - right_x
                    right_error_y = target_y - right_y
                    right_error_xy = np.hypot(right_error_x, right_error_y)

                    lr_error = (right_error_xy + left_error_xy) / 2.0
                    lr_error_max = lr_error.max()
                    lr_error_min = lr_error.min()
                    lr_error_mean = lr_error.mean()
                    lr_error_std = np.std(lr_error)
                    min_error = min(min_error, lr_error_min)
                    max_error = max(max_error, lr_error_max)
                    summed_error += lr_error_mean
                    point_count += 1.0
                else:
                    eye_x = good_samples_in_period[:]['eye_x']
                    eye_y = good_samples_in_period[:]['eye_y']
                    error_x = target_x - eye_x
                    error_y = target_y - eye_y
                    error_xy = np.hypot(error_x, error_y)
                    lr_error = error_xy
                    lr_error_max = lr_error.max()
                    lr_error_min = lr_error.min()
                    lr_error_mean = lr_error.mean()
                    lr_error_std = np.std(lr_error)
                    min_error = min(min_error, lr_error_min)
                    max_error = max(max_error, lr_error_max)
                    summed_error += lr_error_mean
                    point_count += 1.0

                position_results2['calculation_status'] = 'PASSED'
                position_results2['min_error'] = lr_error_min
                position_results2['max_error'] = lr_error_max
                position_results2['mean_error'] = lr_error_mean
                position_results2['stdev_error'] = lr_error_std
            for k, v in position_results2.items():
                self.io.sendMessageEvent('{}: {}'.format(k, v), 'VALIDATION')
                position_results[k] = v
            results['position_results'].append(position_results)
            self.io.sendMessageEvent('Done Target Position Results : {0}'.format(pindex), 'VALIDATION')

        unit_type = self.win.units
        if self.results_in_degrees:
            unit_type = 'degree'
        mean_error = summed_error / point_count
        err_results = dict(reporting_unit_type=unit_type, min_error=min_error, max_error=max_error,
                           mean_error=mean_error, passed=results['positions_failed_processing'] == 0,
                           positions_failed_processing=results['positions_failed_processing'])

        for k, v in err_results.items():
            self.io.sendMessageEvent('{}: {}'.format(k, v), 'VALIDATION')
            results[k] = v

        self.io.sendMessageEvent('Validation Report Complete', 'VALIDATION')

        self._validation_results = results
        return self._validation_results

    def _generateImageName(self):
        import datetime
        file_name = 'validation_' + datetime.datetime.now().strftime('%d_%m_%Y_%H_%M') + '.png'
        #if self.save_results_screen:
        #    return normjoin(self.save_results_screen, file_name)
        rootScriptPath = os.path.dirname(sys.argv[0])
        return normjoin(rootScriptPath, file_name)

    def drawResultScreen(self):
        """
        Draw validation results screen.
        :return:
        """

        results = self.getValidationResults()

        for tp in self.positions.getPositions():
            self.targetsequence.target.pos = tp
            self.targetsequence.target.draw()

        title_txt = 'Validation Results\nMin: %.4f, Max: %.4f,' \
                    ' Mean %.4f (%s units)' % (results['min_error'], results['max_error'],
                                               results['mean_error'], results['reporting_unit_type'])
        title_stim = visual.TextStim(self.win, text=title_txt, height=24, pos=(0.0, (self.win.size[1] / 2.0) * .95),
                                     color=self.text_color, colorSpace=self.text_color_space, units='pix', antialias=True,
                                     anchorVert='center', anchorHoriz='center', wrapWidth=self.win.size[0] * .8)
        title_stim.draw()

        exit_text = visual.TextStim(self.win, text='Press SPACE to continue.', opacity=1.0, units='pix', height=None,
                                    pos=(0.0, -(self.win.size[1] / 2.0) * .95), color=self.text_color, colorSpace=self.text_color_space,
                                    antialias=True, bold=True, anchorVert='center', anchorHoriz='center',
                                    wrapWidth=self.win.size[0] * .8)
        exit_text.draw()

        color_list = pl.cm.tab20b(np.linspace(0, 1, (len(results['position_results']))))
        # draw eye samples
        ci = 0
        for position_results in results['position_results']:
            color = color_list[ci] * 2.0 - 1.0
            utype = 'pix'
            target_x, target_y = position_results['target_position']

            sample_gfx_radius = deg2pix(0.33, self.win.monitor, correctFlat=False)
            if self.results_in_degrees:
                sample_gfx_radius = 0.33
                utype='deg'
            sample_gfx = visual.Circle(self.win, radius=sample_gfx_radius, fillColor=color, lineColor=[1, 1, 1],
                                       lineWidth=1, edges=64, units=utype, colorSpace='rgb', opacity=0.66,
                                       interpolate=True, autoLog=False)

            if position_results['calculation_status'] == 'FAILED':
                position_txt = "Failed"
                txt_bold = True
                position_txt_color = "red"
            else:
                samples = position_results['sample_from_filter_stages']['used_samples']
                binoc_sample_types = [EventConstants.BINOCULAR_EYE_SAMPLE, EventConstants.GAZEPOINT_SAMPLE]
                if self.targetsequence.sample_type in binoc_sample_types:
                    gaze_x = (samples[:]['left_eye_x'] + samples[:]['right_eye_x']) / 2.0
                    gaze_y = (samples[:]['left_eye_y'] + samples[:]['right_eye_y']) / 2.0
                else:
                    gaze_x = samples[:]['eye_x']
                    gaze_y = samples[:]['eye_y']

                for i in range(len(gaze_x)):
                    if self.results_in_degrees:
                        g_pos = gaze_x[i], gaze_y[i]
                    else:
                        g_pos = toPix(self.win, gaze_x[i], gaze_y[i])
                        g_pos = g_pos[0][0], g_pos[1][0]
                    sample_gfx.pos = g_pos
                    sample_gfx.draw()
                txt_bold = False
                position_txt = "Gaze Error:\nMin: %.4f\nMax: %.4f\n" \
                               "Avg: %.4f\nStdev: %.4f" % (position_results['min_error'],
                                                           position_results['max_error'],
                                                           position_results['mean_error'],
                                                           position_results['stdev_error'])
                position_txt_color = "green"

            if self.targetsequence.display_gaze:
                text_pix_pos = toPix(self.win, target_x, target_y)
                text_pix_pos = text_pix_pos[0][0], text_pix_pos[1][0]
                target_text_stim = visual.TextStim(self.win, text=position_txt, units='pix', pos=text_pix_pos,
                                                   height=21, color=position_txt_color, antialias=True,
                                                   bold=txt_bold, anchorVert='center', anchorHoriz='center')
                target_text_stim.draw()
            ci += 1


class ValidationTargetRenderer:
    TARGET_STATIONARY = 1
    TARGET_MOVING = 2
    TARGET_EXPANDING = 4
    TARGET_CONTRACTING = 8
    # Experiment Message text field types and tokens
    message_types = dict(BEGIN_SEQUENCE=('BEGIN_SEQUENCE', '', int),
                         DONE_SEQUENCE=('DONE_SEQUENCE', '', int),
                         NEXT_POS_TRIG=('NEXT_POS_TRIG', '', int, float),
                         START_DRAW=('START_DRAW', ',', int, float, float, float, float),
                         SYNCTIME=('SYNCTIME', ',', int, float, float, float, float),
                         EXPAND_SIZE=('EXPAND_SIZE', '', float, float),
                         CONTRACT_SIZE=('CONTRACT_SIZE', '', float, float),
                         POS_UPDATE=('POS_UPDATE', ',', float, float),
                         TARGET_POS=('TARGET_POS', ',', float, float))
    max_msg_type_length = max([len(s) for s in message_types.keys()])
    binocular_sample_message_element = [('targ_pos_ix', int),
                                        ('last_msg_time', np.float64),
                                        ('last_msg_type', np.str, max_msg_type_length),
                                        ('next_msg_time', np.float64),
                                        ('next_msg_type', np.str, max_msg_type_length),
                                        ('targ_pos_x', np.float64),
                                        ('targ_pos_y', np.float64),
                                        ('targ_state', int),
                                        ('eye_time', np.float64),
                                        ('eye_status', int),
                                        ('left_eye_x', np.float64),
                                        ('left_eye_y', np.float64),
                                        ('left_pupil_size', np.float64),
                                        ('right_eye_x', np.float64),
                                        ('right_eye_y', np.float64),
                                        ('right_pupil_size', np.float64)]
    monocular_sample_message_element = [('targ_pos_ix', int),
                                        ('last_msg_time', np.float64),
                                        ('last_msg_type', np.str, max_msg_type_length),
                                        ('next_msg_time', np.float64),
                                        ('next_msg_type', np.str, max_msg_type_length),
                                        ('targ_pos_x', np.float64),
                                        ('targ_pos_y', np.float64),
                                        ('targ_state', int),
                                        ('eye_time', np.float64),
                                        ('eye_status', int),
                                        ('eye_x', np.float64),
                                        ('eye_y', np.float64),
                                        ('pupil_size', np.float64)]

    def __init__(self, win, target, positions, storeeventsfor=[], triggers=None, msgcategory='',
                 io=None, terminate_key='escape', gaze_cursor_key='g', gaze_cursor=None,
                 color_space=None, unit_type=None):
        """
        ValidationTargetRenderer is an internal class used by `ValidationProcedure`.

        psychopy.iohub.client.eyetracker.validation.Trigger based classes are used
        to define the criteria used to start displaying the next target position graphics.
        By providing a set of DeviceEventTriggers, complex criteria for
        target position pacing can be defined.

        iohub devices can be provided in the storeeventsfor keyword argument.
        Events which occur during each target position presentation period are
        stored and are available at the end of the display() period, grouped by
        position index and device event types.

        :param win:
        :param target:
        :param positions:
        :param storeeventsfor:
        :param triggers:
        :param msgcategory:
        :param io:
        """
        self.terminate_key = terminate_key
        self.gaze_cursor_key = gaze_cursor_key
        self.display_gaze = False
        self.gaze_cursor = None
        if isinstance(gaze_cursor, (str, list, tuple)):
            gc_size = deg2pix(3.0, win.monitor, correctFlat=False)
            self.gaze_cursor = visual.GratingStim(win, tex=None, mask='gauss', pos=(0, 0), size=(gc_size, gc_size),
                                                  color=gaze_cursor, colorSpace=color_space, units='pix', opacity=0.8)
        elif gaze_cursor and hasattr(gaze_cursor, 'setPos'):
            self.gaze_cursor = gaze_cursor
        else:
            raise ValueError("Gaze Cursor must be a color value or visual stim type.")
        self._terminate_requested = False
        self.win = proxy(win)
        self.target = target
        self.positions = positions
        self.storeevents = storeeventsfor
        self.msgcategory = msgcategory

        if io is None:
            io = ioHubConnection.getActiveConnection()
        self.io = io
        self._keyboard = self.io.devices.keyboard

        # If storeevents is True, targetdata will be a list of dict's.
        # Each dict, among other things, contains all ioHub events that occurred
        # from when a target was first presented at a position, to when the
        # the wait period completed for that position.
        #
        self.targetdata = []
        self.triggers = triggers

    def _draw(self):
        """
        Draw the target stim.
        """
        self.target.draw()
        if self.gaze_cursor and self.display_gaze:
            gpos = self.io.devices.tracker.getLastGazePosition()
            valid_gaze_pos = isinstance(gpos, (tuple, list))
            if valid_gaze_pos:
                pix_pos = toPix(self.win, *gpos)
                pix_pos = pix_pos[0][0], pix_pos[1][0]
                self.gaze_cursor.setPos(pix_pos)
                self.gaze_cursor.draw()

    def _animateTarget(self, topos, frompos, **kwargs):
        """
        Any logic related to drawing the target at the new screen position,
        including any intermediate animation effects, is done here.

        Return the flip time when the target was first drawn at the newpos
        location.
        """
        io = self.io

        # Target position animation phase
        animate_position = kwargs.get('enable')
        targetdelay = kwargs.get('targetdelay')
        if frompos is not None:
            if animate_position:
                start_time = getTime()
                while (getTime() - start_time) <= targetdelay:
                    t = (getTime()-start_time) / targetdelay
                    v1 = frompos
                    v2 = topos
                    t = 60.0 * ((1.0 / 10.0) * t ** 5 - (1.0 / 4.0) * t ** 4 + (1.0 / 6.0) * t ** 3)
                    moveTo = ((1.0 - t) * v1[0] + t * v2[0], (1.0 - t) * v1[1] + t * v2[1])
                    self.target.pos = moveTo
                    self._draw()
                    fliptime = self.win.flip()
                    io.sendMessageEvent('POS_UPDATE %.4f,%.4f' % (moveTo[0], moveTo[1]), self.msgcategory,
                                        sec_time=fliptime)
                    self._addDeviceEvents()
                    if self._terminate_requested:
                        return 0
            else:
                # No target animation, so just show cleared screen
                # for targetdelay seconds
                fliptime = self.win.flip(clearBuffer=True)
                while getTime() < fliptime+targetdelay:
                    self._addDeviceEvents()
                    if self._terminate_requested:
                        return 0

        self.target.pos = topos
        self._draw()
        fliptime = self.win.flip()
        io.sendMessageEvent('TARGET_POS %.4f,%.4f' % (topos[0], topos[1]), self.msgcategory, sec_time=fliptime)
        self._addDeviceEvents()

        # Target expand / contract phase
        expandedscale = kwargs.get('expandedscale')
        targetduration = kwargs.get('targetduration')
        contract_target = kwargs.get('contracttarget')
        initialradius = self.target.radius

        expand_duration = None
        contract_duration = None
        if contract_target and expandedscale and expandedscale > 1.0:
            # both expand and contract
            expand_duration = contract_duration = targetduration / 2
        elif contract_target:
            # contract only
            expand_duration = None
            contract_duration = targetduration
        elif expandedscale and expandedscale > 1.0:
            # contract only
            expand_duration = targetduration
            contract_duration = None

        if expand_duration:
            expandedradius = self.target.radius * expandedscale
            starttime = getTime()
            expandedtime = fliptime + expand_duration
            while fliptime < expandedtime:
                mu = (fliptime - starttime) / expand_duration
                cradius = initialradius * (1.0 - mu) + expandedradius * mu
                self.target.radius = cradius
                self._draw()
                fliptime = self.win.flip()
                io.sendMessageEvent('EXPAND_SIZE %.4f %.4f' % (cradius, initialradius), self.msgcategory,
                                    sec_time=fliptime)
                self._addDeviceEvents()
                if self._terminate_requested:
                    return 0

        if contract_duration:
            starttime = getTime()
            contractedtime = fliptime + contract_duration
            start_radius = self.target.radius
            try:
                stop_radius = self.target.innerRadius
            except:
                stop_radius = start_radius/2
                print("Warning: validation target has no .innerRadius property.")
            while fliptime < contractedtime:
                mu = (fliptime - starttime) / contract_duration
                cradius = start_radius * (1.0 - mu) + stop_radius * mu
                self.target.radius = cradius
                self._draw()
                fliptime = self.win.flip()
                io.sendMessageEvent('CONTRACT_SIZE %.4f %.4f' % (cradius, initialradius), self.msgcategory,
                                    sec_time=fliptime)
                self._addDeviceEvents()
                if self._terminate_requested:
                    return 0
        return fliptime

    def moveTo(self, topos, frompos, **kwargs):
        """
        Indicates that the target should be moved frompos to topos.

        If a PositionGrid has been provided, moveTo should not be called
        directly. Instead, use the display() method to start the full
        target position presentation sequence.
        """
        io = self.io
        fpx, fpy = -1, -1
        if frompos is not None:
            fpx, fpy = frompos[0], frompos[1]
        io.sendMessageEvent('START_DRAW %d %.4f,%.4f %.4f,%.4f' % (self.positions.posIndex, fpx, fpy, topos[0],
                                                                   topos[1]), self.msgcategory)

        fliptime = self._animateTarget(topos, frompos, **kwargs)
        io.sendMessageEvent('SYNCTIME %d %.4f,%.4f %.4f,%.4f' % (self.positions.posIndex, fpx, fpy, topos[0], topos[1]),
                            self.msgcategory, sec_time=fliptime)

        # wait for trigger to fire
        last_pump_time = fliptime
        trig_fired = self._hasTriggerFired(start_time=fliptime)
        while not trig_fired:
            if getTime() - last_pump_time >= 0.250:
                win32MessagePump()
                last_pump_time = getTime()

            if self.display_gaze:
                self._draw()
                self.win.flip()
            else:
                sleep(0.001)

            if self._checkForTerminate():
                return
            self._checkForToggleGaze()
            trig_fired = self._hasTriggerFired(start_time=fliptime)

    def _hasTriggerFired(self, **kwargs):
        """
        Used internally to know when one of the triggers has occurred and
        the target should move to the next target position.
        """
        # wait for trigger to fire
        triggered = None
        for trig in self.triggers:
            if trig.triggered(**kwargs):
                triggered = trig
            self._addDeviceEvents(trig.clearEventHistory(True))
            if triggered:
                break

        if triggered:
            # by default, assume it was a timer trigger,so use 255 as 'event type'
            event_type_id = 255
            trig_evt = triggered.getTriggeringEvent()
            if hasattr(trig_evt, 'type'):
                # actually it was a device event trigger
                event_type_id = trig_evt.type
            # get time trigger of trigger event
            event_time = triggered.getTriggeringTime()
            self.io.sendMessageEvent('NEXT_POS_TRIG %d %.3f' % (event_type_id, event_time), self.msgcategory)
            for trig in self.triggers:
                trig.resetTrigger()
        return triggered

    def _initTargetData(self, frompos, topos):
        """
        Internally used to create the data structure used to store position
        information and events which occurred during each target position
        period.
        """
        if self.storeevents:
            deviceevents = {}
            for device in self.storeevents:
                deviceevents[device] = []
        self.targetdata.append(dict(frompos=frompos, topos=topos, events=deviceevents))

    def _addDeviceEvents(self, device_event_dict={}):
        if self._checkForTerminate():
            return
        self._checkForToggleGaze()
        dev_event_buffer = self.targetdata[-1]['events']
        for dev, dev_events in dev_event_buffer.items():
            if dev in device_event_dict:
                dev_events.extend(device_event_dict[dev])
            else:
                dev_events.extend(dev.getEvents())

    def _checkForTerminate(self):
        keys = self._keyboard.getEvents(EventConstants.KEYBOARD_PRESS, clearEvents=False)
        for k in keys:
            if k.key == self.terminate_key:
                self._terminate_requested = True
                break
        return self._terminate_requested

    def _checkForToggleGaze(self):
        keys = self._keyboard.getEvents(EventConstants.KEYBOARD_PRESS, clearEvents=False)
        for k in keys:
            if k.key == self.gaze_cursor_key:
                # get (clear) the event so it does not trigger multiple times.
                self._keyboard.getEvents(EventConstants.KEYBOARD_PRESS, clearEvents=True)
                self.display_gaze = not self.display_gaze
                self._draw()
                self.win.flip()
                return self.display_gaze
        return self.display_gaze

    def display(self, **kwargs):
        """
        Display the target at each point in the position grid, performing
        target animation if requested. The target then holds position until one
        of the specified triggers occurs, resulting in the target moving to the
        next position in the positiongrid.

        To setup target animation between grid positions, the following keyword
        arguments are supported. If an option is not specified, the animation
        related to it is not performed.



        Note that target expansion and contraction change the target stimulus
        outer diameter only. The edge thickness and central dot radius do not
        change.

        When this method returns, the target has been displayed at all
        positions. Data collected for each position period can be accessed via
        the targetdata attribute.
        """
        del self.targetdata[:]
        prevpos = None

        io = self.io
        io.clearEvents('all')
        io.sendMessageEvent('BEGIN_SEQUENCE {0}'.format(len(self.positions.positions)), self.msgcategory)
        turn_rec_off = []
        for d in self.storeevents:
            if not d.isReportingEvents():
                d.enableEventReporting(True)
                turn_rec_off.append(d)

        sleep(0.025)
        initialsize=self.target.radius
        for pos in self.positions:
            self._initTargetData(prevpos, pos)
            self._addDeviceEvents()
            if self._terminate_requested:
                break
            self.target.radius = initialsize
            self.moveTo(pos, prevpos, **kwargs)
            prevpos = pos
            self._addDeviceEvents()
            if self._terminate_requested:
                break
        self.target.radius = initialsize
        for d in turn_rec_off:
            d.enableEventReporting(False)

        if self._terminate_requested:
            io.sendMessageEvent('VALIDATION TERMINATED BY USER', self.msgcategory)
            return False

        io.sendMessageEvent('DONE_SEQUENCE {0}'.format(len(self.positions.positions)), self.msgcategory)
        sleep(0.025)
        self._addDeviceEvents()
        io.clearEvents('all')
        return True

    def _processMessageEvents(self):
        self.target_pos_msgs = []
        self.saved_pos_samples = []
        for pd in self.targetdata:
            events = pd.get('events')

            # create a dict of device labels as keys, device events as value
            devlabel_events = {}
            for k, v in events.items():
                devlabel_events[k.getName()] = v

            samples = devlabel_events.get('tracker', [])
            # remove any eyetracker events that are not samples
            samples = [s for s in samples if s.type in (EventConstants.BINOCULAR_EYE_SAMPLE,
                                                        EventConstants.MONOCULAR_EYE_SAMPLE,
                                                        EventConstants.GAZEPOINT_SAMPLE)]
            self.saved_pos_samples.append(samples)

            self.sample_type = self.saved_pos_samples[0][0].type
            if self.sample_type == EventConstants.MONOCULAR_EYE_SAMPLE:
                self.sample_msg_dtype = self.monocular_sample_message_element
            else:
                self.sample_msg_dtype = self.binocular_sample_message_element
            messages = devlabel_events.get('experiment', [])
            msg_lists = []
            for m in messages:
                temp = m.text.strip().split()
                msg_type = self.message_types.get(temp[0])
                if msg_type:
                    current_msg = [m.time, m.category]
                    if msg_type[1] == ',':
                        for t in temp:
                            current_msg.extend(t.split(','))
                    else:
                        current_msg.extend(temp)

                    for mi, dtype in enumerate(msg_type[2:]):
                        current_msg[mi + 3] = dtype(current_msg[mi + 3])

                    msg_lists.append(current_msg)

            if msg_lists[0][2] == 'NEXT_POS_TRIG':
                # handle case where the trigger msg from the previous target
                # message was not read until the start of the next pos.
                # In which case, move msg to end of previous targ pos msgs
                npm = msg_lists.pop(0)
                self.target_pos_msgs[-1].append(npm)

            self.target_pos_msgs.append(msg_lists)

        for i in range(len(self.target_pos_msgs)):
            self.target_pos_msgs[i] = np.asarray(self.target_pos_msgs[i], dtype=object)

        return self.target_pos_msgs

    def getSampleMessageData(self):
        """
        Return a list of numpy ndarrays, each containing joined eye sample
        and previous / next experiment message data for the sample's time.
        """
        # preprocess message events
        self._processMessageEvents()

        # inline func to return sample field array based on sample namedtup
        def getSampleData(s):
            sampledata = [s.time, s.status]
            binoc_sample_types = [EventConstants.BINOCULAR_EYE_SAMPLE, EventConstants.GAZEPOINT_SAMPLE]
            if s.type in binoc_sample_types:
                sampledata.extend((s.left_gaze_x, s.left_gaze_y, s.left_pupil_measure1,
                                   s.right_gaze_x, s.right_gaze_y, s.right_pupil_measure1))
                return sampledata

            sampledata.extend((s.gaze_x, s.gaze_y, s.pupil_measure1))
            return sampledata

        current_target_pos = -1.0, -1.0
        current_targ_state = 0
        target_pos_samples = []
        for pindex, samples in enumerate(self.saved_pos_samples):
            last_msg, messages = self.target_pos_msgs[pindex][0], self.target_pos_msgs[pindex][1:]
            samplesforposition = []
            pos_sample_count = len(samples)
            si = 0
            for current_msg in messages:
                last_msg_time = last_msg[0]
                last_msg_type = last_msg[2]
                if last_msg_type == 'START_DRAW':
                    if not current_targ_state & self.TARGET_STATIONARY:
                        current_targ_state += self.TARGET_STATIONARY
                    current_targ_state -= current_targ_state & self.TARGET_MOVING
                    current_targ_state -= current_targ_state & self.TARGET_EXPANDING
                    current_targ_state -= current_targ_state & self.TARGET_CONTRACTING
                elif last_msg_type == 'EXPAND_SIZE':
                    if not current_targ_state & self.TARGET_EXPANDING:
                        current_targ_state += self.TARGET_EXPANDING
                    current_targ_state -= current_targ_state & self.TARGET_CONTRACTING
                elif last_msg_type == 'CONTRACT_SIZE':
                    if not current_targ_state & self.TARGET_CONTRACTING:
                        current_targ_state += self.TARGET_CONTRACTING
                    current_targ_state -= current_targ_state & self.TARGET_EXPANDING
                elif last_msg_type == 'TARGET_POS':
                    current_target_pos = float(last_msg[3]), float(last_msg[4])
                    current_targ_state -= current_targ_state & self.TARGET_MOVING
                    if not current_targ_state & self.TARGET_STATIONARY:
                        current_targ_state += self.TARGET_STATIONARY
                elif last_msg_type == 'POS_UPDATE':
                    current_target_pos = float(last_msg[3]), float(last_msg[4])
                    if not current_targ_state & self.TARGET_MOVING:
                        current_targ_state += self.TARGET_MOVING
                    current_targ_state -= current_targ_state & self.TARGET_STATIONARY
                elif last_msg_type == 'SYNCTIME':
                    if not current_targ_state & self.TARGET_STATIONARY:
                        current_targ_state += self.TARGET_STATIONARY
                    current_targ_state -= current_targ_state & self.TARGET_MOVING
                    current_targ_state -= current_targ_state & self.TARGET_EXPANDING
                    current_targ_state -= current_targ_state & self.TARGET_CONTRACTING
                    current_target_pos = float(last_msg[6]), float(last_msg[7])

                while si < pos_sample_count:
                    sample = samples[si]
                    if last_msg_time <= sample.time < current_msg[0]:
                        sarray = [pindex, last_msg_time, last_msg_type,
                                  current_msg[0], current_msg[2],
                                  current_target_pos[0], current_target_pos[1],
                                  current_targ_state]
                        sarray.extend(getSampleData(sample))
                        sndarray = np.asarray(tuple(sarray), dtype=self.sample_msg_dtype)
                        samplesforposition.append(sndarray)
                        si += 1
                    elif sample.time >= current_msg[0]:
                        break
                    else:
                        si += 1
                last_msg = current_msg

            possamples = np.asanyarray(samplesforposition)
            target_pos_samples.append(possamples)

        # So we now have a list len == number target positions. Each element
        # of the list is a list of all eye sample / message data for a
        # target position. Each element of the data list for a single target
        # position is itself a list that that contains combined info about
        # an eye sample and message info valid for when the sample time was.
        return np.asanyarray(target_pos_samples, dtype=object)


def toPix(win, x, y):
    """Returns the stim's position in pixels,
    based on its pos, units, and win.
    """
    try:
        xy = np.zeros((len(x), 2))
    except TypeError:
        xy = np.zeros((1, 2))

    xy[:, 0] = x
    xy[:, 1] = y
    r = convertToPix(np.asarray((0, 0)), xy, win.units, win)
    return r[:, 0], r[:, 1]


def toDeg(win, x, y):
    try:
        xy = np.zeros((len(x), 2))
    except TypeError:
        xy = np.zeros((1, 2))
    xy[:, 0] = x
    xy[:, 1] = y
    r = pix2deg(xy, win.monitor, correctFlat=False)
    return r[:, 0], r[:, 1]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Class for more control over the mouse,
including the pointer graphic and bounding box."""

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).




import os

import psychopy  # so we can get the __path__
from psychopy import event, logging, visual, layout

from psychopy.visual.basevisual import MinimalStim

import numpy


class CustomMouse(MinimalStim):
    """Class for more control over the mouse,
    including the pointer graphic and bounding box.

    Seems to work with pyglet or pygame. Not completely tested.

    Known limitations:

    * only norm units are working
    * getRel() always returns [0,0]
    * mouseMoved() is always False; maybe due to `self.mouse.visible == False` -> held at [0,0]
    * no idea if clickReset() works

    Author: Jeremy Gray, 2011
    """

    def __init__(self, win, newPos=None, visible=True,
                 leftLimit=None, topLimit=None,
                 rightLimit=None, bottomLimit=None,
                 showLimitBox=False, clickOnUp=False,
                 pointer=None, name=None, autoLog=None):
        """Class for customizing the appearance and behavior of the mouse.

        Use a custom mouse for extra control over the pointer appearance
        and function. It's probably slower to render than the regular
        system mouse. Create your `visual.Window` before creating a
        CustomMouse.

        Parameters
        ----------

        win : required, `visual.Window`
            the window to which this mouse is attached
        visible : **True** or False
            makes the mouse invisible if necessary
        newPos : **None** or [x,y]
            gives the mouse a particular starting position
        leftLimit :
            left edge of a virtual box within which the mouse can move
        topLimit :
            top edge of virtual box
        rightLimit :
            right edge of virtual box
        bottomLimit :
            lower edge of virtual box
        showLimitBox : default is False
            display the boundary within which the mouse can move.
        pointer :
            The visual display item to use as the pointer;
            must have .draw() and setPos() methods. If your item has
            .setOpacity(), you can alter the mouse's opacity.
        clickOnUp : when to count a mouse click as having occurred
            default is False, record a click when the mouse is first
            pressed down. True means record a click when the mouse
            button is released.

        """
        # what local vars are defined (these are the init params) for use by
        # __repr__
        self._initParams = dir()
        self._initParams.remove('self')
        super(CustomMouse, self).__init__(name=name, autoLog=False)
        self.autoLog = False  # set properly at end of init

        self.win = win
        self.mouse = event.Mouse(win=self.win)

        # maybe inheriting from Mouse would be easier? its not that simple
        self.getRel = self.mouse.getRel
        self.getWheelRel = self.mouse.getWheelRel
        self.mouseMoved = self.mouse.mouseMoved  # FAILS
        self.mouseMoveTime = self.mouse.mouseMoveTime
        self.getPressed = self.mouse.getPressed
        self.clickReset = self.mouse.clickReset  # ???
        self._pix2windowUnits = self.mouse._pix2windowUnits
        self._windowUnits2pix = self.mouse._windowUnits2pix

        # the graphic to use as the 'mouse' icon (pointer)
        if pointer:
            self.pointer = pointer
        else:
            self.pointer = vm = visual.ShapeStim(
                win,
                vertices=[
                    [-0.5, 0.5],
                    [-0.5, -0.35],
                    [-0.3, -0.15],
                    [-0.1, -0.5],
                    [0.025, -0.425],
                    [-0.175, -0.1],
                    [0.1, -0.1],
                    [-0.5, 0.5],
                ],
                fillColor="white",
                lineColor="black",
                lineWidth=1,
                anchor="top left",
                size=layout.Size((20, 20), 'pix', win),
            )
        self.mouse.setVisible(False)  # hide the actual (system) mouse
        self.visible = visible  # the custom (virtual) mouse

        self.leftLimit = self.rightLimit = None
        self.topLimit = self.bottomLimit = None
        self.setLimit(leftLimit=leftLimit, topLimit=topLimit,
                      rightLimit=rightLimit, bottomLimit=bottomLimit)
        self.showLimitBox = showLimitBox

        self.lastPos = None
        self.prevPos = None
        if newPos is not None:
            self.lastPos = newPos
        else:
            self.lastPos = self.mouse.getPos()

        # for counting clicks:
        self.clickOnUp = clickOnUp
        self.wasDown = False  # state of mouse 1 frame prior to current frame
        self.clicks = 0  # how many mouse clicks since last reset
        self.clickButton = 0  # which button to count clicks for; 0 = left

        # set autoLog now that params have been initialised
        wantLog = autoLog is None and self.win.autoLog
        self.__dict__['autoLog'] = autoLog or wantLog
        if self.autoLog:
            logging.exp("Created %s = %s" % (self.name, str(self)))

    def _setPos(self, pos=None):
        """internal mouse position management. setting a position here
        leads to the virtual mouse being out of alignment with the hardware
        mouse, which leads to an 'invisible wall' effect for the mouse.
        """
        if pos is None:
            pos = self.getPos()
        else:
            self.lastPos = pos
        self.pointer.setPos(pos)

    def setPos(self, pos):
        """Not implemented yet. Place the mouse at a specific position.
        """
        raise NotImplementedError('setPos is not available for custom mouse')

    def getPos(self):
        """Returns the mouse's current position. Influenced
        by changes in .getRel(), constrained to be in its virtual box.
        """
        dx, dy = self.getRel()
        x = min(max(self.lastPos[0] + dx, self.leftLimit), self.rightLimit)
        y = min(max(self.lastPos[1] + dy, self.bottomLimit), self.topLimit)
        self.lastPos = numpy.array([x, y])
        return self.lastPos

    def draw(self):
        """Draw mouse (if it's visible), show the limit box, update
        the click count.
        """
        self._setPos()
        if self.showLimitBox:
            self.box.draw()
        if self.visible:
            self.pointer.draw()
        isDownNow = self.getPressed()[self.clickButton]
        if self.clickOnUp:
            if self.wasDown and not isDownNow:  # newly up
                self.clicks += 1
        else:
            if not self.wasDown and isDownNow:  # newly down
                self.clicks += 1
        self.wasDown = isDownNow

    def getClicks(self):
        """Return the number of clicks since the last reset"""
        return self.clicks

    def resetClicks(self):
        """Set click count to zero"""
        self.clicks = 0

    def getVisible(self):
        """Return the mouse's visibility state"""
        return self.visible

    def setVisible(self, visible):
        """Make the mouse visible or not (pyglet or pygame)."""
        self.visible = visible

    def setPointer(self, pointer):
        """Set the visual item to be drawn as the mouse pointer."""
        if hasattr(pointer, 'draw') and hasattr(pointer, 'setPos'):
            self.pointer = pointer
        else:
            raise AttributeError("need .draw() and .setPos() methods"
                                 " in pointer")

    def setLimit(self, leftLimit=None, topLimit=None,
                 rightLimit=None, bottomLimit=None):
        """Set the mouse's bounding box by specifying the edges.
        """
        if type(leftLimit) in (int, float):
            self.leftLimit = leftLimit
        elif self.leftLimit is None:
            self.leftLimit = -1
            if self.win.units == 'pix':
                self.leftLimit = self.win.size[0] / -2.
        if type(rightLimit) in (int, float):
            self.rightLimit = rightLimit
        elif self.rightLimit is None:
            self.rightLimit = .99
            if self.win.units == 'pix':
                self.rightLimit = self.win.size[0] / 2.0 - 5
        if type(topLimit) in (int, float):
            self.topLimit = topLimit
        elif self.topLimit is None:
            self.topLimit = 1
            if self.win.units == 'pix':
                self.topLimit = self.win.size[1] / 2.0
        if type(bottomLimit) in (int, float):
            self.bottomLimit = bottomLimit
        elif self.bottomLimit is None:
            self.bottomLimit = -0.98
            if self.win.units == 'pix':
                self.bottomLimit = self.win.size[1] / -2.0 + 10

        self.box = psychopy.visual.ShapeStim(
            self.win, vertices=[[self.leftLimit, self.topLimit],
                                [self.rightLimit, self.topLimit],
                                [self.rightLimit, self.bottomLimit],
                                [self.leftLimit, self.bottomLimit],
                                [self.leftLimit, self.topLimit]],
            opacity=0.35, autoLog=False)

        # avoid accumulated relative-offsets producing a different effective
        # limit:
        self.mouse.setVisible(True)
        self.lastPos = self.mouse.getPos()  # hardware mouse's position
        self.mouse.setVisible(False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2022 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Functions and classes related to coordinate system conversion
"""
import numpy
from numpy import radians


def cart2pol(x, y, units='deg'):
    """Convert from cartesian to polar coordinates.

    :usage:

        theta, radius = cart2pol(x, y, units='deg')

    units refers to the units (rad or deg) for theta that should be returned
    """
    radius = numpy.hypot(x, y)
    theta = numpy.arctan2(y, x)
    if units in ('deg', 'degs'):
        theta = theta * 180 / numpy.pi
    return theta, radius


def pol2cart(theta, radius, units='deg'):
    """Convert from polar to cartesian coordinates.

    usage::

        x,y = pol2cart(theta, radius, units='deg')

    """
    if units in ('deg', 'degs'):
        theta = theta * numpy.pi / 180.0
    xx = radius * numpy.cos(theta)
    yy = radius * numpy.sin(theta)

    return xx, yy


def cart2sph(z, y, x):
    """Convert from cartesian coordinates (x,y,z) to spherical (elevation,
    azimuth, radius). Output is in degrees.

    usage:
        array3xN[el,az,rad] = cart2sph(array3xN[x,y,z])
        OR
        elevation, azimuth, radius = cart2sph(x,y,z)

        If working in DKL space, z = Luminance, y = S and x = LM
    """
    width = len(z)

    elevation = numpy.empty([width, width])
    radius = numpy.empty([width, width])
    azimuth = numpy.empty([width, width])

    radius = numpy.sqrt(x**2 + y**2 + z**2)
    azimuth = numpy.arctan2(y, x)
    # Calculating the elevation from x,y up
    elevation = numpy.arctan2(z, numpy.sqrt(x**2 + y**2))

    # convert azimuth and elevation angles into degrees
    azimuth *= 180.0 / numpy.pi
    elevation *= 180.0 / numpy.pi

    sphere = numpy.array([elevation, azimuth, radius])
    sphere = numpy.rollaxis(sphere, 0, 3)

    return sphere


def sph2cart(*args):
    """Convert from spherical coordinates (elevation, azimuth, radius)
    to cartesian (x,y,z).

    usage:
        array3xN[x,y,z] = sph2cart(array3xN[el,az,rad])
        OR
        x,y,z = sph2cart(elev, azim, radius)
    """
    if len(args) == 1:  # received an Nx3 array
        elev = args[0][0, :]
        azim = args[0][1, :]
        radius = args[0][2, :]
        returnAsArray = True
    elif len(args) == 3:
        elev = args[0]
        azim = args[1]
        radius = args[2]
        returnAsArray = False

    z = radius * numpy.sin(radians(elev))
    x = radius * numpy.cos(radians(elev)) * numpy.cos(radians(azim))
    y = radius * numpy.cos(radians(elev)) * numpy.sin(radians(azim))
    if returnAsArray:
        return numpy.asarray([x, y, z])
    else:
        return x, y, z



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
PsychoPy's log module is mostly a simple wrapper of the python logging module.
It allows messages to be sent from any location (any library or your script) and
written to files according to the importance level of the message and the minimum level
that the log file (or console) is set to receive. You can have multiple log files each receiving
different levels of input.

The importance levels are
    40: ERROR
    35: DATA
    30: WARNING
    25: DATA
    22: EXP
    20: INFO
    10: DEBUG
So setting to DEBUG level will include all possible messages, setting to ERROR will include only the absolutely essential messages.
"""

from psychopy import logging, core, visual

globalClock = core.Clock()  # if this isn't provided the log times will reflect secs since python started
logging.setDefaultClock(globalClock)

logging.console.setLevel(logging.DEBUG)  # receive nearly all messages
logDat = logging.LogFile('logLastRun.log',
    filemode='w',  # if you set this to 'a' it will append instead of overwriting
    level=logging.WARNING)  # errors, data and warnings will be sent to this logfile

# the following will go to any files with the appropriate minimum level set
logging.info('Something fairly unimportant')
logging.data('Something about our data. Data is likely very important!')
logging.warning('Handy while building your experiment - highlights possible flaws in code/design')
logging.error("You might have done something that PsychoPy can't handle! But hopefully this gives you some idea what.")

# some things should be logged timestamped on the next video frame
# For instance the time of a stimulus appearing is related to the flip:
win = visual.Window([400, 400])
for n in range(5):
    win.logOnFlip('frame %i occured' %n, level=logging.EXP)
    if n in [2, 4]:
        win.logOnFlip('an even frame occured', level=logging.EXP)
    win.flip()

# LogFiles can also simply receive direct input from the write() method
# messages using write() will be sent immediately, and are often not
# in correct chronological order with logged messages
logDat.write("Testing\n\n")

win.close()
core.quit()

# The contents of this file are in the public domain.
